/* 
 * Copyright 2007 TKK/Netlab
 * Released under GPLv3. See LICENSE.txt for details. 
 */
package core;
import movement.ExternalMovement;
import movement.MovementModel;
import ui.DTNSimTextUI;
import gui.DTNSimGUI;

/**
 * Simulator's main class 
 */
public class DTNSim {
	/** If this option ({@value}) is given to program, batch mode and
	 * Text UI are used*/
	public static final String BATCH_MODE_FLAG = "-b";
	
	/**
	 * Starts the user interface with given arguments.
	 * If first argument is {@link #BATCH_MODE_FLAG}, the text UI is started.
	 * Next argument is the settings file for the simulation run (if any).
	 * For text UI, also number of runs can be given after the setting file 
	 * name. Different settings from run arrays are used for different runs (see
	 * {@link Settings#setRunIndex(int)}). For GUI mode, the number after 
	 * settings file (if given) is the run index to use for that run.
	 * @param args Command line arguments
	 */
	public static void main(String[] args) {
		boolean batchMode = false;
		String confFile = null;
		int nrofRuns = 1;
		// set US locale to parse decimals in consistent way
		java.util.Locale.setDefault(java.util.Locale.US);
		
		if (args.length > 0) {
			if (args[0].equals(BATCH_MODE_FLAG)) {
				batchMode = true;
				if (args.length > 1) {
					confFile = args[1];
				}
				if (args.length > 2) {
					nrofRuns = parseNrofRuns(args[2]);
				}
			}
			else {
				confFile = args[0];
				if (args.length > 1) { // use specific run index
					int runIndex = parseNrofRuns(args[1]);
					Settings.setRunIndex(runIndex);
				}
			}
		}
		
		initSettings(confFile);
		
		if (batchMode) {
			long startTime = System.currentTimeMillis();
			for (int i=0; i<nrofRuns; i++) {
				if (nrofRuns > 1) {
					print("Run " + (i+1) + "/" + nrofRuns);
				}
				Settings.setRunIndex(i);
				resetForNextRun();
				new DTNSimTextUI().start();
			}
			double duration = (System.currentTimeMillis() - startTime)/1000.0;
			print("---\nAll done in " + String.format("%.2f", duration) + "s");
		}
		else {
			new DTNSimGUI().start();
		}
	}
	
	/**
	 * Initializes Settings
	 * @param confFile Where to read additional settings (in addition to
	 * default settings) or null if there are no additional settings
	 */
	private static void initSettings(String confFile) {
		try {
			Settings.init(confFile);
		}
		catch (SettingsError er) {
			System.err.println("Failed to load settings: " + er);
			System.err.println("Caught at " + er.getStackTrace()[0]);
			System.exit(-1);
		}
	}
	
	/**
	 * Some classes need resetting their static fields before new run
	 */
	private static void resetForNextRun() {
		/* TODO: Write a resetting-registering method to this class and 
		 * register the classes that need resetting in their code and reset
		 * them all here. */
		Message.reset();
		SimClock.reset();
		NetworkLayer.reset();
		MovementModel.reset();
		ExternalMovement.reset();
	}
	
	/**
	 * Parses the number of runs from command line argument
	 * @param arg The argument to parse
	 * @return The value as integer
	 */
	private static int parseNrofRuns(String arg) {
		int val = 1;
		
		try {
			val = Integer.parseInt(arg);
		} catch (NumberFormatException e) {
			System.err.println("Invalid numeric argument '" + arg + "' for" +
					" number of runs");
			System.exit(-1);
		}
		
		if (val < 1) {
			val = 1;
		}
		
		return val;
	}
	
	/**
	 * Prints text to stdout
	 * @param txt Text to print
	 */
	private static void print(String txt) {
		System.out.println(txt);
	}
}
