/* 
 * Copyright 2007 TKK/Netlab
 * Released under GPLv3. See LICENSE.txt for details. 
 */
package movement;

import java.util.Random;

import core.*;
import movement.distribution.NormalDistribution;
import movement.distribution.ParetoDistribution;
import movement.distribution.RandomDistribution;
import movement.distribution.UniformDistribution;

/**
 * <P>Superclass for all movement models. All subclasses must contain at least a 
 * constructor with one {@link Settings} parameter and also a copy-constructor.
 * They must also implement the {@link #replicate()} method, which should return
 * an instance of the movement model class with same parameters as the object
 * (immutable fields can be shared, but mutable fields must be copied).</P>
 * <P>To make a new movement model do something usefull, also at least
 * {@link #getInitialLocation()} and {@link #getPath()} are worthwhile to 
 * override.</P>
 */
public abstract class MovementModel {
	/** node's speed CSV (min, max) -setting id ({@value})*/
	public static final String SPEED = "speed";
	/** node's wait time CSV (min, max) -setting id ({@value})*/
	public static final String WAIT_TIME = "waitTime";
	
        public static final String SPEEDDIST = "speedDist";
        public static final String WAITTIMEDIST = "waitTimeDist";
        
	/** default setting for speed distribution */
	public static final double[] DEF_SPEEDS = {1,1};
	/** default setting for wait time distribution */
	public static final double[] DEF_WAIT_TIMES = {0,0};
	
	/** MovementModel namespace (where world size and rng seed settings 
	 * are looked from ({@value})*/
	public static final String MOVEMENT_MODEL_NS = "MovementModel";
	/** world's size CSV (width, height) -setting id ({@value})*/
	public static final String WORLD_SIZE = "worldSize";
	/** movement models' rng seed -setting id ({@value})*/
	public static final String RNG_SEED = "rngSeed";
	
	/** common rng for all movement models in the simulation */
	protected static Random rng; 
	
	private ActivenessHandler ah;
		
        private RandomDistribution randistSpeed, randistWT;
        
	private double speedParamA;
	private double speedParamB;
	private double wtParamA;
	private double wtParamB;

	private int maxX;
	private int maxY;

	// static initialization of all movement models' random number generator
	static {
		reset();
	}
	
	/**
	 * Checks that the minimum setting is not bigger than the maximum and
	 * that both are positive
	 * @param name Name of the setting
	 * @param min The minimum setting
	 * @param max The maximum setting
	 */
	private static void checkMinAndMaxSetting(String name,
		double min, double max) {
		if (min < 0 || max < 0) {
			throw new SimError("MovementModel." + name + " (in Settings)" + 
					" has a value less than zero ("+min+", "+max+")");
		}
		if (min > max) {
			throw new SimError("MovementModel." + name + " (in Settings)" + 
					" min is bigger than max ("+min+", "+max+")");
		}
	}
	
	/** 
	 * Empty constructor for testing purposes.
	 */
	public MovementModel() {
		super();
	}
	
	/**
	 * Creates a new MovementModel based on a Settings object's settings.
	 * @param settings The Settings object where the settings are read from
	 */
	public MovementModel(Settings settings) {
		double[] speeds;
		double[] times;
		
		ah = new ActivenessHandler(settings);
		
		if (settings.contains(SPEED)) {
			speeds = settings.getCsvDoubles(SPEED, 2);
		}
		else {
			speeds = DEF_SPEEDS;
		}
		
		speedParamA = speeds[0];
		speedParamB = speeds[1];
		//checkMinAndMaxSetting(SPEED,minSpeed,maxSpeed);

                if(settings.contains(SPEEDDIST)) {
			String speedDist = settings.getSetting(SPEEDDIST);
                        if (speedDist.compareTo("normal") == 0)
                            randistSpeed = new NormalDistribution(speedParamA, speedParamB, "normal", rng);
                        else if (speedDist.compareTo("uniform") == 0)
                            randistSpeed = new UniformDistribution(speedParamA, speedParamB, "uniform", rng);
                        else if (speedDist.compareTo("exponential") == 0)
                            randistSpeed = null;
                        else if (speedDist.compareTo("pareto") == 0)
                            randistSpeed = new ParetoDistribution(speedParamA, speedParamB, "pareto", rng);                           
		}
                else
                    randistSpeed = new UniformDistribution(speedParamA, speedParamB, "uniform", rng);
                
		if(settings.contains(WAIT_TIME)) {
			times = settings.getCsvDoubles(WAIT_TIME, 2);
		}
		else {
			times = DEF_WAIT_TIMES;
		}
		
		wtParamA = times[0];
		wtParamB = times[1];
		//checkMinAndMaxSetting(WAIT_TIME,minWaitTime,maxWaitTime);
		
                if(settings.contains(WAITTIMEDIST)) {
			String wtDist = settings.getSetting(SPEEDDIST);
                        if (wtDist.compareTo("normal") == 0)
                            randistWT = new NormalDistribution(wtParamA, wtParamB, "normal", rng);
                        else if (wtDist.compareTo("uniform") == 0)
                            randistWT = new UniformDistribution(wtParamA, wtParamB, "uniform", rng);
                        else if (wtDist.compareTo("exponential") == 0)
                            randistWT = null;
                        else if (wtDist.compareTo("pareto") == 0)
                            randistWT = new ParetoDistribution(wtParamA, wtParamB, "pareto", rng);
		}
                else
                    randistWT = new UniformDistribution(wtParamA, wtParamB, "uniform", rng);
                
		settings.setNameSpace(MOVEMENT_MODEL_NS);
		int [] worldSize = settings.getCsvInts(WORLD_SIZE,2);
		this.maxX = worldSize[0];
		this.maxY = worldSize[1];

		settings.restoreNameSpace();
                
                DisasterMovement.instantiate();
	}
	
	/**
	 * Copyconstructor. Creates a new MovementModel based on the given
	 * prototype.
	 * @param mm The MovementModel prototype to base the new object to 
	 */
	public MovementModel(MovementModel mm) {
		this.speedParamA = mm.speedParamA;
		this.speedParamB = mm.speedParamB;
		this.wtParamA = mm.wtParamA;
		this.wtParamA = mm.wtParamB;
                this.randistSpeed = mm.randistSpeed.replicate();
                this.randistWT = mm.randistWT.replicate();
		this.maxX = mm.maxX;
		this.maxY = mm.maxY;
		this.ah = mm.ah;
	}
		
	/**
	 * Returns the largest X coordinate value this model uses
	 * @return Maximum of X coordinate values
	 */
	public int getMaxX() {
		return this.maxX;
	}

	/**
	 * Returns the largest Y coordinate value this model uses
	 * @return Maximum of Y coordinate values
	 */
	public int getMaxY() {
		return this.maxY;
	}

	
	/**
	 * Generates and returns a speed value between min and max of the 
	 * {@link #WAIT_TIME} setting.
	 * @return A new speed between min and max values 
	 */
	protected double generateSpeed() {
		if (rng == null) {
			return 1;
		}
		//return (maxSpeed - minSpeed) * rng.nextDouble() + minSpeed;
                return randistSpeed.nextValue();
	}
	
	/**
	 * Generates and returns a suitable waiting time at the end of a path.
	 * (i.e. random variable whose value is between min and max of the 
	 * {@link #WAIT_TIME} setting).
	 * @return The time as a double
	 */
	protected double generateWaitTime() {
		if (rng == null) {
			return 0;
		}
		//return (maxWaitTime - minWaitTime) * rng.nextDouble() + 
		//	minWaitTime;
                
                return randistWT.nextValue();
	}

	/**
	 * Returns a new path by this movement model or null if no new path could
	 * be constructed at the moment (node should wait where it is). A new
	 * path should not be requested before the destination of the previous
	 * path has been reached.
	 * @return A new path or null
	 */
	public abstract Path getPath();
	
	/**
	 * Returns a new initial placement for a node
	 * @return The initial coordinates for a node
	 */
	public abstract Coord getInitialLocation();
	
	/**
	 * Returns true if this node is active at the moment (false if not)
	 * @return true if this node is active (false if not)
	 */
	public boolean isActive() {
		return ah.isActive();
	}
		
	/**
	 * Returns a sim time when the next path is available. This implementation
	 * returns a random time in future that is {@link #WAIT_TIME} from now.
	 * @return The sim time when node should ask the next time for a path
	 */
	public double nextPathAvailable() {
		return SimClock.getTime() + generateWaitTime();
	}
	
	/**
	 * Returns simply the name of the movement model class
	 * @return the name of the movement model class
	 */
        @Override
	public String toString() {
		return this.getClass().getSimpleName();
	}
	
	/**
	 * Creates a replicate of the movement model.
	 * @return A new movement model with the same settings as this model
	 */
	public abstract MovementModel replicate();
	
	/**
	 * Resets all static fields to default values
	 */
	public static void reset() {
		Settings s = new Settings(MOVEMENT_MODEL_NS);
		if (s.contains(RNG_SEED)) {
			int seed = s.getInt(RNG_SEED);
			rng = new Random(seed);
		}
		else {
			rng = new Random(0);
		}
                
                DisasterMovement.reset();
	}
	
}
