/* 
 * Copyright 2007 TKK/Netlab
 * Released under GPLv3. See LICENSE.txt for details. 
 */
package report;

import java.util.*;

import core.DTNHost;
import core.Message;
import core.MessageListener;
import core.SimClock;
import java.util.ArrayList;

/**
 * Report for generating different kind of total statistics about message
 * relaying performance. 
 * <P><strong>Note:</strong> if some statistics could not be created (e.g.
 * overhead ratio if no messages were delivered) "NaN" is reported for
 * double values and zero for integer median(s).
 */
public class MessageStatsReport extends Report implements MessageListener {
	/** Returnvalue for values that could not be calculated */
	public static final String NAN = "NaN";
	
	private Map<String, Double> creationTimes;
	private List<Double> latencies;
	private List<Integer> hopCounts;
	private List<Double> msgBufferTime;
	private List<Double> rtt; // round trip times
    
    private List<Double> propagationTime;
	
	private int nrofDropped;
	private int nrofRemoved;
	private int nrofStarted;
	private int nrofAborted;
	private int nrofRelayed;
	private int nrofCreated;
	private int nrofResponseReqCreated;
	private int nrofResponseDelivered;
	private int nrofDelivered;
	
    private int phase;
	/**
	 * Constructor.
	 */
	public MessageStatsReport() {
		init();
        phase = 0;
	}

	@Override
	protected void init() {
		super.init();
		this.creationTimes = new HashMap<String, Double>();
		this.latencies = new ArrayList<Double>();
		this.msgBufferTime = new ArrayList<Double>();
		this.hopCounts = new ArrayList<Integer>();
		this.rtt = new ArrayList<Double>();
		
        this.propagationTime = new ArrayList<Double>();
        
		this.nrofDropped = 0;
		this.nrofRemoved = 0;
		this.nrofStarted = 0;
		this.nrofAborted = 0;
		this.nrofRelayed = 0;
		this.nrofCreated = 0;
		this.nrofResponseReqCreated = 0;
		this.nrofResponseDelivered = 0;
		this.nrofDelivered = 0;
	}

	
	public void messageDeleted(Message m, DTNHost where, boolean dropped) {
		if (m.getId().startsWith("summary"))
                    return;
                if (dropped) {
			this.nrofDropped++;
		}
		else {
			this.nrofRemoved++;
		}
		
		this.msgBufferTime.add(getSimTime() - m.getReceiveTime());
	}

	
	public void messageTransferAborted(Message m, DTNHost from, DTNHost to) {
                if (m.getId().startsWith("summary"))
                    return;
		this.nrofAborted++;
	}

	
	public void messageTransferred(Message m, DTNHost from, DTNHost to,
			boolean finalTarget) {
                if (m.getId().startsWith("summary"))        
                    return;
		this.nrofRelayed++;
		if (finalTarget) {
			this.latencies.add(getSimTime() - 
				this.creationTimes.get(m.getId()) );
			this.nrofDelivered++;
			this.hopCounts.add(m.getHops().size() - 1);
		
            /* <yusufsarwar> */                        
            if (m.getId().compareTo("M22") == 0)
            {
                System.out.println("here it is");
            }
            String str = getScenarioName() + " " + m.getId() + " C: " + m.getCreationTime() + " S: " + m.getStartTime() + " D: " + SimClock.getTime() + " " + m.getHops();
            System.out.println(str);            
            /* </yusufsarwar> */
            
			if (m.isResponse()) {
				this.rtt.add(getSimTime() -	m.getRequest().getCreationTime());
				this.nrofResponseDelivered++;
			}                        
            this.propagationTime.add(SimClock.getTime() - m.getStartTime());
		}
        
        if (SimClock.getTime() > phase * 3600)
        {
            done();
            phase++;
        }
	}


	public void newMessage(Message m) {		            
                if (m.getId().startsWith("summary"))
                    return;
                this.creationTimes.put(m.getId(), getSimTime());
		this.nrofCreated++;
		if (m.getResponseSize() > 0) {
			this.nrofResponseReqCreated++;
		}
	}
	
	public void messageTransferStarted(Message m, DTNHost from, DTNHost to) {
            if (m.getId().startsWith("summary"))
                    return;	
            this.nrofStarted++;
	}
	
	@Override
	public void done() {
		write("Message stats for scenario " + getScenarioName() + 
				"\nsim_time: " + format(getSimTime()));
		double deliveryProb = 0; // delivery probability
		double responseProb = 0; // request-response success probability
		double overHead = Double.NaN;	// overhead ratio
		
		if (this.nrofCreated > 0) {
			deliveryProb = (1.0 * this.nrofDelivered) / this.nrofCreated;
		}
		if (this.nrofDelivered > 0) {
			//overHead = (1.0 * (this.nrofRelayed - this.nrofDelivered)) /
			//	this.nrofDelivered;
                        overHead = 1.0 * this.nrofRelayed / this.nrofCreated;
		}
		if (this.nrofResponseReqCreated > 0) {
			responseProb = (1.0* this.nrofResponseDelivered) / 
				this.nrofResponseReqCreated;
		}
		
		String statsText = "created: " + this.nrofCreated + 
			"\nstarted: " + this.nrofStarted + 
			"\nrelayed: " + this.nrofRelayed +
			"\naborted: " + this.nrofAborted + 
			"\ndropped: " + this.nrofDropped +
			"\nremoved: " + this.nrofRemoved +
			"\ndelivered: " + this.nrofDelivered +
			"\ndelivery_prob: " + format(deliveryProb) +
			"\nresponse_prob: " + format(responseProb) + 
			"\noverhead_ratio: " + format(overHead) + 
			"\nlatency_avg: " + getAverage(this.latencies) +
			"\nlatency_med: " + getMedian(this.latencies) + 
            "\npropagation_avg: " + getAverage(this.propagationTime) +
			"\npropagation_med: " + getMedian(this.propagationTime) +             
			"\nhopcount_avg: " + getIntAverage(this.hopCounts) +
			"\nhopcount_med: " + getIntMedian(this.hopCounts) + 
			"\nbuffertime_avg: " + getAverage(this.msgBufferTime) +
			"\nbuffertime_med: " + getMedian(this.msgBufferTime) +
			"\nrtt_avg: " + getAverage(this.rtt) +
			"\nrtt_med: " + getMedian(this.rtt)
			;	
                
		write(statsText);
        
        statsText = "created: " + this.nrofCreated + 
			//" started: " + this.nrofStarted + 			
		//	" aborted: " + this.nrofAborted + 
		//	" dropped: " + this.nrofDropped +
		//	" removed: " + this.nrofRemoved +
		//	" delivered: " + this.nrofDelivered +
			" delivery_prob: " + format(deliveryProb) +
	//		" response_prob: " + format(responseProb) + 
			" overhead_ratio: " + format(overHead) + 
			" latency_avg: " + getAverage(this.latencies) +
			" latency_med: " + getMedian(this.latencies) +              
                        " relayed: " + this.nrofRelayed +
			" hopcount_avg: " + getIntAverage(this.hopCounts) +
			//" hopcount_med: " + getIntMedian(this.hopCounts) + 
			//" buffertime_avg: " + getAverage(this.msgBufferTime) +
			//" buffertime_med: " + getMedian(this.msgBufferTime) +
	//		" rtt_avg: " + getAverage(this.rtt) +
	//		" rtt_med: " + getMedian(this.rtt) +
			"";	
        System.out.println("Scene: " + getScenarioName() + " hour " + SimClock.getTime() / 3600 + " " + statsText);
		super.done();
	}

	/**
	 * Returns the average of double values stored in a List or "NaN" for
	 * empty lists.
	 * @param values The list of double values
	 * @return average of double values stored in the List in a formatted String 
	 */
	private String getAverage(List<Double> values) {
		double sum = 0;
		if (values.size() == 0) {
			return NAN;
		}

		for (double dValue : values) {
			sum += dValue;
		}
		
		return format(sum / values.size());
	}

	/**
	 * Returns the average of integer values stored in a List
	 * @param values The list of values
	 * @return average of integer values stored in the List or "NaN" for
	 * empty lists.
	 */
	private String getIntAverage(List<Integer> values) {
		List<Double> dValues = new ArrayList<Double>(values.size());
		for (int i : values) {
			dValues.add((double)i);
		}	
		return getAverage(dValues);
	}
	
	/**
	 * Returns the median of double values stored in a List
	 * @param values The list of double values
	 * @return median of double values stored in the List or "NaN" for
	 * empty lists.
	 */
	private String getMedian(List<Double> values) {
		if (values.size() == 0) {
			return NAN;
		}
		
		Collections.sort(values);
		return format(values.get(values.size()/2));
	}
	
	/**
	 * Returns the median of integer values stored in a List
	 * @param values The list of values
	 * @return median of integer values stored in the List or 0 for
	 * empty lists.
	 */
	private int getIntMedian(List<Integer> values) {
		if (values.size() == 0) {
			return 0;
		}
		
		Collections.sort(values);
		return values.get(values.size()/2);
	}             
}
