/* 
 * Copyright 2007 TKK/Netlab
 * Released under GPLv3. See LICENSE.txt for details. 
 */
package report;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;

import core.Settings;
import core.SimClock;
import core.SimError;
import core.SimScenario;

/**
 * Abstract superclass for all reports. All settings defined in this class
 * can be used for all Report classes. Some reports don't implement intervalled
 * reports ({@link #INTERVAL_SETTING}) and will ignore that setting.
 */
public abstract class Report {
	/** The interval (simulated seconds) of creating new settings files 
	 * -setting id ({@value}) */
	public static final String INTERVAL_SETTING = "interval";
	/** The output file path of the report -setting id ({@value})*/
	public static final String OUTPUT_SETTING = "output";
	/** Precision of formatted double values - setting id ({@value}). 
	 * Defines the amount of decimals shown in formatted double values.
	 * Default value is {@value #DEF_PRECISION}. */
	public static final String PRECISION_SETTING = "precision";
	/** Default precision of formatted double values */
	public static final int DEF_PRECISION = 4;
	/** The default output directory of reports (can be overriden per report 
	 * with {@link Report#OUTPUT_SETTING}) -setting id ({@value})*/
	public static final String REPORTDIR_SETTING = "Report.reportDir";
	
	/** Suffix of report files without explicit output */
	public static final String OUT_SUFFIX = ".txt";
	/** Suffix for reports that are created on n second intervals */
	public static final String INTERVALLED_FORMAT ="%04d" + OUT_SUFFIX; 
	/** The print writer used to write output. See {@link #write(String)} */
	protected PrintWriter out;
	private String prefix = "";
	private int precision;
	
	private int lastOutputSuffix;
	private double outputInterval;
	private double lastReportTime;
	private String outFileName;
	private String scenarioName;
	
	/**
	 * Constructor.
	 * Looks for a className.output setting in the Settings and
	 * if such is found, uses that as the output file name. Otherwise
	 * scenarioname_classname.txt is used as the file name.
	 */
	public Report(){
		this.lastOutputSuffix = 0;
		this.outputInterval = -1;

		Settings settings = new Settings();
		scenarioName = settings.valueFillString(settings.getSetting(
				SimScenario.SCENARIO_NS + "." +	SimScenario.NAME_S));

		// switch to Report-specific namespace
		settings.setNameSpace(this.getClass().getSimpleName());
				
		if (settings.contains(INTERVAL_SETTING)) {
			outputInterval = settings.getDouble(INTERVAL_SETTING);
		}
		
		if (settings.contains(PRECISION_SETTING)) {
			precision = settings.getInt(PRECISION_SETTING);
			if (precision < 0) {
				precision = 0;
			}
		}
		else {
			precision = DEF_PRECISION;
		}
		
		if (settings.contains(OUTPUT_SETTING)) {
			outFileName = settings.getSetting(OUTPUT_SETTING);
			// fill value place holders in the name
			outFileName = settings.valueFillString(outFileName);
		}
		else {
			// no output name define -> construct one from report class' name
			settings.setNameSpace(null);
			String outDir = settings.getSetting(REPORTDIR_SETTING);
			if (!outDir.endsWith("/")) {
				outDir += "/";	// make sure dir ends with directory delimiter
			}
			outFileName = outDir + scenarioName + 
				"_" + this.getClass().getSimpleName();
			if (outputInterval == -1) {
				outFileName += OUT_SUFFIX; // no intervalled reports
			}
			
		}
				
		checkDirExistence(outFileName);
	}

	/**
	 * Checks that a directory for a file exists or creates the directory
	 * if it didn't exist.
	 * @param outFileName Name of the file
	 */
	private void checkDirExistence(String outFileName) {
		File outFile = new File(outFileName);
		File outDir = outFile.getParentFile();
	
		if (outDir != null && !outDir.exists()) {
			if (!outDir.mkdir()) {
				throw new SimError("Couldn't create report directory '" + 
						outDir.getAbsolutePath()+"'");
			}
		}
	}
	
	/**
	 * Intializes the report output. Method is called in the beginning of
	 * every new report file. Subclasses must call this method first in their
	 * own implementations of init().
	 */
	protected void init() {
		this.lastReportTime = getSimTime();
		
		if (outputInterval > 0) {
			createSuffixedOutput(outFileName);
		}
		else {
			createOutput(outFileName);
		}
	}
	
	/**
	 * Creates a new output file
	 * @param outFileName Name (&path) of the file to create
	 */
	private void createOutput(String outFileName) {
		try {
			this.out = new PrintWriter(new FileWriter(outFileName));
		} catch (IOException e) {
			throw new SimError("Couldn't open file '" + outFileName + 
					"' for report output\n" + e.getMessage(), e);
		}		
	}
	
	/**
	 * Creates a number-suffixed output file with increasing number suffix
	 * @param outFileName Prefix of the output file's name
	 */
	private void createSuffixedOutput(String outFileName) {
		String suffix = String.format(INTERVALLED_FORMAT,
				this.lastOutputSuffix);
		createOutput(outFileName+suffix);
		this.lastOutputSuffix++;
	}
	
	/**
	 * This method should be called before every new (complete) event the
	 * report logs. If the report has no meaningful use for multiple reports,
	 * the call can be omitted (then only single output file will be generated) 
	 */
	protected void newEvent() {
		if (this.outputInterval <= 0) {
			return;
		}
		
		if (getSimTime() > this.lastReportTime + this.outputInterval) {
			done(); // finalize the old file
			init(); // init the new file
		}
	}
	
	/**
	 * Writes a line to report using defined prefix and {@link #out} writer.
	 * @param txt Line to write
	 * @see #setPrefix(String)
	 */
	protected void write(String txt) {
		if (out == null) {
			init();
		}
		out.println(prefix + txt);
	}
	
	/**
	 * Formats a double value according to current precision setting (see
	 * {@link #PRECISION_SETTING}) and returns it in a string.
	 * @param value The value to format
	 * @return Formatted value in a string
	 */
	protected String format(double value) {
		return String.format("%." + precision + "f", value);
	}
	
	/**
	 * Sets a prefix that will be inserted before every line in the report
	 * @param txt Text to use as the prefix
	 */
	protected void setPrefix(String txt) {
		this.prefix = txt;
	}
	
	/**
	 * Returns the name of the scenario as read from the settings
	 * @return the name of the scenario as read from the settings
	 */
	protected String getScenarioName() {
		return this.scenarioName;
	}
	
	/**
	 * Returns the current simulation time from the SimClock
	 * @return the current simulation time from the SimClock
	 */
	protected double getSimTime() {
		return SimClock.getTime();
	}
	
	/**
	 * Called when the simulation is done, user requested
	 * premature termination or intervalled report generating decided
	 * that it's time for the next report.
	 */
	public void done() {
		out.close();
	}
		
}
