/* 
 * Copyright 2007 TKK/Netlab
 * Released under GPLv3. See LICENSE.txt for details. 
 */
package report;

import java.util.List;

import core.DTNHost;
import core.UpdateListener;

/**
 * Report for total amount of contact times among hosts. Reports how long all
 * nodes have been in contact with some other node. Supports 
 * {@link ContactTimesReport#GRANULARITY} setting. If update interval is
 * smaller than 1.0 seconds, time stamps may start to drift. Reported values
 * still correspond to reported times.
 */
public class TotalContactTimeReport extends ContactTimesReport implements
		UpdateListener {
	
	/** The header of every report file */
	public static final String HEADER = "# time totalContactTime";
	/** cumulative contact times of all disconnected contacts */
	private double oldContactTimes;
	/** sim time of last report writing */
	private double lastWrite;
	/** last reported time count (to suppress duplicates) */
	private double lastReportedTime;
	
	public void init() {
		super.init();
		write(HEADER);
		this.oldContactTimes = 0;
		this.lastReportedTime = 0;
		this.lastWrite = getSimTime();
	}

	@Override
	public void hostsDisconnected(DTNHost host1, DTNHost host2) {
		newEvent();
		ConnectionInfo ci = removeConnection(host1, host2);
		
		assert ci != null : "No active connection of "  + 
			host1 +	" and " + host2 + " found in " + this.connections;

		oldContactTimes += ci.getConnectionTime();		
	}

	/**
	 * Reports total contact time if more time than defined with setting
	 * {@value ContactTimesReport#GRANULARITY} has passed. Method is called
	 * on every update cycle.
	 */
	public void updated(List<DTNHost> hosts) {
		double simTime = getSimTime();
		if (simTime - lastWrite < granularity) {
			return; // shouldn't report yet
		}
		lastWrite = simTime;
		
		// count also the times for connections that are stil up
		double othersTime = 0;
		for (ConnectionInfo oth : this.connections.values()) {
			othersTime += oth.getConnectionTime();
		}
		
		double totalTime = oldContactTimes + othersTime;
		
		if (lastReportedTime == totalTime) {
			return; // don't report duplicate times
		}
		
		write(format(simTime) + " " + format(totalTime));
		lastReportedTime = totalTime;
	}
}
