/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package routing;

/**
 *
 * @author Yusuf Sarwar
 */
/* 
 * Copyright 2007 TKK/Netlab
 * Released under GPLv3. See LICENSE.txt for details. 
 */

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import routing.contact.Contact;
import routing.contact.NeighborEntry;
import routing.contact.PeerEvent;
import routing.contact.PeerEventEntry;
import routing.contact.RoutingTable;
import core.Connection;
import core.DTNHost;
import core.Message;
import core.Settings;
import core.SimClock;
import core.Tuple;
/**
 * Implementation of contact router as described in 
 * <I>Probabilistic routing in intermittently connected networks</I> by
 * Anders Lindgren et al.
 */
public class ContactRouter extends ActiveRouter {
	/** contact router's setting namespace ({@value})*/ 
	public static final String contact_NS = "ContactRouter";

	private static boolean debug;
	private static final boolean keepCopy = false;
	private static final boolean keepContactList = false;

	/** Contact lists */
	private Map<DTNHost, NeighborEntry> visitedNodes;
	private Map<DTNHost, NeighborEntry> neighbors;        
        private DTNHost lastVisitedNeighbor;

	private int noOfNeighbors, noOfTotalNodes;
	private double contactDelays[][], contactVariance[][], contactQs[][], routeDelays[][], routeVariance[][];
	private int routeHopCount[][];
	private int contactCounts[][];
	private int messageCopies;

	private Map<String, List<DTNHost>> contactList;
        
	private List<String> deliveredMessages;

	private final int DELAY_MAX = 6000;
	private final static int visitCountThreshold = 5;
	private final double eta = 1.0;
	private final double gamma = 0.75;
	private boolean newConnection = false;
	private boolean toBeForwarded = false;

	/**
	 * Constructor. Creates a new message router based on the settings in
	 * the given Settings object.
	 * @param s The settings object
	 */
	public ContactRouter(Settings s) {
		super(s);        

		s.setNameSpace(contact_NS);		
		noOfTotalNodes = s.getInt("nrofHosts");
		noOfNeighbors = noOfTotalNodes;

		if (s.contains("debug")) {
			debug = s.getBoolean("debug");
		}
		else {
			debug = true;
		}

		if (s.contains("messageCopies"))
			messageCopies = s.getInt("messageCopies");
		else
			messageCopies = 10;
		initContacts();
	}

	/**
	 * Copyconstructor.
	 * @param r The router prototype where setting values are copied from
	 */
	protected ContactRouter(ContactRouter r) {
		super(r);      
		noOfTotalNodes = r.noOfTotalNodes;
		noOfNeighbors = r.noOfNeighbors;
		messageCopies = r.messageCopies;
		initContacts();
	}

	/**
	 * Initializes predictability hash
	 */
	private void initContacts() {           
		this.visitedNodes = new HashMap<DTNHost, NeighborEntry>();
		this.neighbors = new HashMap<DTNHost, NeighborEntry>();
		//this.routingTables = new HashMap<Contact, RoutingTable>();            
		//forwardingEvents =  new HashMap<Message, DTNHost>();            

		this.contactDelays = new double[noOfNeighbors][noOfNeighbors];
		this.contactVariance = new double[noOfNeighbors][noOfNeighbors];
		this.contactQs = new double[noOfNeighbors][noOfNeighbors];
		this.contactCounts = new int[noOfNeighbors][noOfNeighbors];

		this.contactList = new HashMap<String, List<DTNHost>>();

		this.routeDelays = new double[noOfNeighbors][noOfTotalNodes];
		this.routeVariance = new double[noOfNeighbors][noOfTotalNodes];
		this.routeHopCount = new int[noOfNeighbors][noOfTotalNodes];

		for(int i = 0; i < noOfNeighbors; i++)
                    for (int j = 0; j < noOfTotalNodes; j++)
                    {
                        routeDelays[i][j] = -1.0;                      
                        routeVariance[i][j] = Double.MAX_VALUE;
                        routeHopCount[i][j] = 0;                                
                    }
		deliveredMessages = new ArrayList<String>();                
	}

	@Override
	public void changedConnection(Connection con) {
		if (con.isUp()) {                
			DTNHost otherHost = con.getOtherNode(getHost());                
			meetANode(otherHost);
			if (visitedNodes.get(otherHost).getVisitedCounter() > visitCountThreshold)
			{                    
				updateContactFor(otherHost);
				updateContacts(otherHost);
                                computeNextNodeForMessages(otherHost);
				newConnection = true;
			}
		}
	}

	private void meetANode(DTNHost host)
	{
		NeighborEntry ne = visitedNodes.get(host);
		if (ne == null)
		{
			visitedNodes.put(host, new NeighborEntry(visitedNodes.size()));
			ne = visitedNodes.get(host);
		}            

		ne.setVisitedCounter(ne.getVisitedCounter() + 1);
		ne.setLastVisited(SimClock.getTime());

		if (keepCopy) {
			ContactRouter otherRouter = (ContactRouter)host.getRouter();
			List<String> list =  otherRouter.getDeliveredMessages();

			if (list.size() > 0)
				for (String m : list) {
					if (!deliveredMessages.contains(m))            
						deliveredMessages.add(m);               
					if (this.getMessage(m) != null)
						deleteMessage(m, false);
				}
		}
	}

        private void computeNextNodeForMessages(DTNHost host)
        {
            for (Message m: getMessageCollection())
            {
                ContactMessage cm = (ContactMessage)m;
                if (cm.toForward.size() == 0)
                {
                    cm.setToForward(findBestNextNode(host, cm.getTo(), m, true));
                    log(cm + " sets next nodes " + cm.getToForward());
                }
            }
        }
	/**
	 * Updates Contact for a host.
	 * <CODE>T(ab, ac) = alpha * T(ab, ac)_old + (1 - alpha) * (ab, ac) </CODE>
	 * @param host The host we just met
	 */
	private void updateContactFor(DTNHost host) 
	{
		if (!neighbors.containsKey(host) && neighbors.size() >= noOfNeighbors) return;

                lastVisitedNeighbor = host;
		/* Find pair-wise meeting contact with this host */
		for (Map.Entry<DTNHost, NeighborEntry> e : neighbors.entrySet())                
		{     
			/* ij -> ik happens i: this node, j: neighbor, k: just visited */
			if (e.getKey().equals(host)) continue; /* not the current node it meets */
			double lastContactTime = SimClock.getTime() - e.getValue().getLastVisited(); 

			setDelay_ij_ik(e.getKey(), host, lastContactTime);
		}            

		NeighborEntry nentry = neighbors.get(host);
		if (nentry != null)
		{
			nentry.setLastVisited(SimClock.getTime());
			nentry.setVisitedCounter(nentry.getVisitedCounter() + 1);
		}
		else
			if (neighbors.size() < noOfNeighbors) 
				neighbors.put(host, new NeighborEntry(SimClock.getTime(), 1, neighbors.size()));

		/* Penalty for not meeting for a while */
                /*
		List<DTNHost> dels = new ArrayList<DTNHost>();
		for(Map.Entry<DTNHost, NeighborEntry> e : neighbors.entrySet())
		{
			DTNHost h1 = e.getKey();
			double mt1 = e.getValue().getLastVisited();		

			if (SimClock.getTime() - mt1 > DELAY_MAX)
			{
				int counter = neighbors.get(h1).getVisitedCounter();
				counter--;
				if (counter < 3)
				{
					dels.add(h1);
				}
				else
				{
					neighbors.get(h1).setVisitedCounter(counter);
					neighbors.get(h1).setLastVisited(SimClock.getTime());
				}
			}
		}            
		for(DTNHost h : dels)
		{
			neighbors.remove(h);
			visitedNodes.get(h).setVisitedCounter(0);
		}
                */ 
	}     
	
        private void updateContacts(DTNHost host) {                                  

		if (!neighbors.containsKey(host)) return;            
		DTNHost hj = host;
		/* for each destination, dest w, compute c(ij-->w) */
		int dest;
		for(dest = 0; dest < noOfTotalNodes; dest++)
		{
			if (dest == getHost().getAddress()) continue;
			setDelay_ij_w(hj, dest);            
		}        
	}

	void showContacts(Map<PeerEvent, PeerEventEntry> map, String name)
	{
		log(name);
		for (Map.Entry<PeerEvent, PeerEventEntry> e : map.entrySet())
			log(e.getKey() + " : " + e.getValue());
	}       

	/**
	 * Schedules messages originated from this very node to some destination 
	 */

	@Override
	public void update() {            
		super.update();
		if (!canStartTransfer() || isTransferring()) {
			return; // nothing to transfer or is currently transferring 
		}

		// try messages that could be delivered to final recipient
		if (exchangeDeliverableMessages() != null) {
			return;
		}
		tryOtherMessages();		
	}

	@Override
	public int receiveMessage(Message m, DTNHost from) {		            
		int ret = super.receiveMessage(m, from);		
		return ret;
	}

	@Override 
	public boolean createNewMessage(DTNHost to, String id, int size,
			int responseSize) {		

		makeRoomForNewMessage(size);		
		ContactMessage cm = new ContactMessage(getHost(), to, id, size);
		cm.setResponseSize(responseSize);
		addToMessages(cm, true);
                
                List<NextNode> nextNodes = null;
                nextNodes = findBestNextNode(lastVisitedNeighbor, to, cm, false);                
                log(getHost() + " sets next nodes for " + cm + " " + nextNodes);
                cm.setToForward(nextNodes);
		return true;
	}

	/**
	 * Tries to send all other messages to all connected hosts ordered by
	 * their delivery probability
	 * @return The return value of {@link #tryMessagesForConnected(List)}
	 */

	private Tuple<Message, Connection> tryOtherMessages() {            
		List<Tuple<Message, Connection>> messages =                     
			new ArrayList<Tuple<Message, Connection>>(); 

		//Collection<Message> msgCollection = getMessageCollection();

		for (Connection con : getConnections()) 
		{
                    DTNHost other = con.getOtherNode(getHost());

                    //if (neighbors.containsKey(other) == false) continue; /* Not a neighbor node */

                    ContactRouter otherRouter = (ContactRouter)other.getRouter();                

                    if (otherRouter.isTransferring()) {
                            continue; // skip hosts that are transferring
                    }                              

                    for (Message m : getMessageCollection())
                    {
                        assert m instanceof ContactMessage;                   

                        // Check if the message is there or visited                    
                        if (m.getHops().contains(other) || otherRouter.getMessage(m.getId()) != null) continue; // the message already there

                        // Check if destination is neighbors

                        //if (neighbors.containsKey(m.getTo()))
                        //        continue;
                        // Deciding the next hop for m

                        ContactMessage cm = (ContactMessage)m;                    

                        if (cm.forwarded) continue; // already forwarded
                        
                        if (cm.getToForward().size() > 0)
                            log(getHost() + " deciding for " + m + " cands " + cm.getToForward() + " Meet " + other);
                        
                        if (cm.getToForward().size() == 0)
                        {
                            if (cm.copiesLeft > 1)
                            {
                                cm.isOnRoute = false;
                                messages.add(new Tuple<Message, Connection>(m, con));
                            }
                        }
                        else 
                        {
                            boolean contains = false;                        
                            for (NextNode n : cm.getToForward())
                                if (n.node.getAddress() == other.getAddress())
                                {
                                    contains = true;
                                    break;                         
                                }
                            if (contains)
                            {
                                // already scheduled to this node
                                log(getHost() + " forwards " + m + " to " + other + " prescheduled " + SimClock.getTime());
                                //cm.forwarded = true;
                                cm.isOnRoute = true;                                
                                messages.add(new Tuple<Message, Connection>(m,con));
                                continue;
                            }                                                                               
                        }
                    }
		}

		if (messages.size() == 0) {
			return null;
		}

		// sort the message-connection tuples                
		Collections.sort(messages, new TupleComparator());
		return tryMessagesForConnected(messages);	// try to send messages
	}      
        
        private class NextNode 
        {
            public DTNHost node;
            public double delay;
            
            public NextNode(DTNHost h, double d)
            {
                node = h;
                delay = d;
            }
            
            public boolean equals(NextNode n)
            {
                return this.node.getAddress() == n.node.getAddress();
            }
            
            public boolean equals(DTNHost n)
            {
                return node.getAddress() == n.getAddress();
            }
            
            @Override
            public String toString()
            {
                return node.toString() + ":" + delay;
            }
        }
        
        private class NextNodeComparator implements Comparator<NextNode>
        {
            public int compare(ContactRouter.NextNode o1, ContactRouter.NextNode o2) {
                if (o1.delay == o2.delay) return 0;
                if (o1.delay > o2.delay) 
                    return 1;
                else
                    return -1;
            }            
        }
        
        private List<NextNode> findBestNextNode(DTNHost other, DTNHost dest, Message m, boolean fromContact)
        {
            double minDelay, delay_ij_ik, delay_ik_w, offset;                         
            List<NextNode> list = new ArrayList<NextNode>();
            
            if (other == null) return null;
            minDelay = Double.MAX_VALUE;           

            if (fromContact == false)
            {
                offset = SimClock.getTime() - neighbors.get(other).getLastVisited();
            }
            else
                offset = 0.0;
            
            log("Computing next hop for " + getHost() + " /w neighbors " + neighbors.size());
            for (Map.Entry<DTNHost, NeighborEntry> e : neighbors.entrySet())
            {
                    DTNHost hk = e.getKey();
                    if (hk.equals(other) || m.getHops().contains(hk)) continue;
                    
                    delay_ij_ik = getDelay_ij_ik(other, hk);
                    //log("Check1 " + delay_ij_ik);
                    if (delay_ij_ik < 0) continue;
                    
                    delay_ik_w = getDelay_ij_w(hk, dest);                                               
                    //log("Check2 " + delay_ik_w);    
                    if (delay_ik_w < 0) continue;

                    double alternateDelay = delay_ij_ik + delay_ik_w - offset;
                    //log("New node added " + hk + " " + alternateDelay);
                    list.add(new NextNode(hk, alternateDelay));
            }
            System.out.println("Next node selection with size " + list.size());
            
            Collections.sort(list, new NextNodeComparator());            
            if (list.size() <= 2)
                return list;
            return list.subList(0, 2);
        }
        
	public double getBestRouteToDestination(DTNHost dest)
	{
		DTNHost prevNode = null;
		double lastTime = 0;

		for(Map.Entry<DTNHost, NeighborEntry> e : neighbors.entrySet())
		{
                    if (lastTime < e.getValue().getLastVisited())
                    {
                            lastTime = e.getValue().getLastVisited();
                            prevNode = e.getKey();
                    }
		}

		double leastDelay = -1.0;
		// DTNHost nextNode = null;

		for (Map.Entry<DTNHost, NeighborEntry> e : neighbors.entrySet())
		{
			// delay = d(jl_->jl) - (now - last) + c(jl ---> dest)             
			DTNHost next = e.getKey();

			if (getDelay_ij_ik(prevNode, next) < 0)
				continue;

			double currDelay =  getDelay_ij_ik(prevNode, next) - SimClock.getTime() + lastTime;
			currDelay = currDelay + getDelay_ij_w(next, dest);

			if (leastDelay < 0 || currDelay < leastDelay)
			{
                            leastDelay = currDelay;
                            // nextNode = next;
			}
		}

		return leastDelay;
	}

	@Override
	public Message messageTransferred(String id, DTNHost from) {

		ContactMessage msg = (ContactMessage)super.messageTransferred(id, from);				
		msg.advanceNextHopIndex();
		if (msg.isOnRoute == false)
			msg.copiesLeft = (int)Math.ceil(msg.copiesLeft/2.0);		
		msg.forwarded = false;
                
                log(getHost() + " Message received from "+ from);
                if (!msg.getTo().equals(getHost()))
                {
                    List<NextNode> nextHops = findBestNextNode(from, msg.getTo(), msg, true);
                    log(msg + " gets next nodes " + nextHops);
                    msg.setToForward(nextHops);
                }
		return msg;
	}  

	@Override
	protected void transferDone(Connection con) {            
		Message m = con.getMessage();            

		DTNHost other = con.getOtherNode(getHost());

		System.out.println("@@" + m.getId() + " " + m.getTo().getAddress() + " " + getHost().getAddress()  + "-" + con.getOtherNode(getHost()).getAddress() + " " + SimClock.getTime());
		log(getHost() + " transfer done " + m + " to " + other);

		Message msg = this.getMessage(m.getId());

		if (msg == null) return;

		assert msg instanceof ContactMessage: "Message type mismatched!!!";

		ContactMessage cmsg = (ContactMessage)msg;
                
                if (cmsg.toForward != null && cmsg.toForward.size() > 0)
                {
                    List<NextNode> list = cmsg.getToForward();
                    for(int i=0; i < list.size(); i++)
                    {
                        if (other.getAddress() == list.get(i).node.getAddress())
                        {
                            list.remove(i);
                            break;
                        }
                    }                    
                    if (list.size()==0)
                        cmsg.forwarded = true;
                    else
                        cmsg.setToForward(list);
                }

		//cmsg.estimatedDelay = cmsg.estimatedDelay - (SimClock.getTime() - cmsg.lastForwarded);            
		//if (cmsg.estimatedDelay < 0)
		//    cmsg.estimatedDelay = 0.0;

		cmsg.lastForwarded = SimClock.getTime();

		if (other.equals(m.getTo()))
		{               
			/* Destination reached at the other end */
			String hops="";                        
			for(DTNHost h : m.getHops())
				hops = hops + " " + h.toString() ;              
			hops = "C: " + m.getCreationTime() + " S: " + ((ContactMessage)m).getStartTime() + " D: " + SimClock.getTime() + " DT: " + (SimClock.getTime()-cmsg.getStartTime()) + " ET: " + cmsg.estimatedDelay;
			System.out.println(m.getId() + "(" + m.getFrom() + "->" + m.getTo() +") " + hops);

			if (keepCopy) 
				deliveredMessages.add(m.getId());

			this.deleteMessage(m.getId(), false);
		}   
		else
		{                
			if (cmsg.isOnRoute == true)                    
				cmsg.advanceNextHopIndex();                
			else
				cmsg.copiesLeft /= 2;

			if (!keepCopy)
				this.deleteMessage(m.getId(), false);                	
		}            
	}

	/**
	 * Comparator for Message-Connection-Tuples that orders the tuples by
	 * their delivery probability by the host on the other side of the 
	 * connection (GRTRMax)
	 */
	private class TupleComparator implements Comparator                 
	<Tuple<Message, Connection>> {

		public int compare(Tuple<Message, Connection> tuple1,
				Tuple<Message, Connection> tuple2) {                

			ContactMessage m1 = (ContactMessage)tuple1.getKey();
			ContactMessage m2 = (ContactMessage)tuple2.getKey();

			return (int)(m1.estimatedDelay - m2.estimatedDelay);
		}            
	}

	@Override
	public RoutingInfo getRoutingInfo() {        

		RoutingInfo top = super.getRoutingInfo();
		RoutingInfo ri = new RoutingInfo(neighbors.size() + 
				" neighors(s)");

		for (Map.Entry<DTNHost, NeighborEntry> e: neighbors.entrySet())
			ri.addMoreInfo(new RoutingInfo(String.format("%s meet %s", e.getKey(), e.getValue())));      
		top.addMoreInfo(ri);

		RoutingInfo ri1 = new RoutingInfo(neighbors.size() + 
				" contacts(s)");

		for(Map.Entry<DTNHost, NeighborEntry> hj : neighbors.entrySet())
                    for(Map.Entry<DTNHost, NeighborEntry> hk : neighbors.entrySet())
                    {
                            int j, k;
                            j = getNeighborIndex(hj.getKey());
                            k = getNeighborIndex(hk.getKey());

                            if (j == k || contactCounts[j][k] == 0) continue;

                            DTNHost hi = getHost();

                            ri1.addMoreInfo(new RoutingInfo(String.format("(%s,%s) -> (%s,%s) = %.3f :: %.3f", hi, hj.getKey(), hi, hk.getKey(), contactDelays[j][k], contactQs[j][k])));      
                    }
		top.addMoreInfo(ri1);

		RoutingInfo ri2 = new RoutingInfo(neighbors.size() + " contacts(s)");

		for(Map.Entry<DTNHost, NeighborEntry> hj : neighbors.entrySet())
                    for(int w = 0; w < noOfTotalNodes; w++)
                    {
                        int j;

                        j = getNeighborIndex(hj.getKey());                                                

                        if (routeDelays[j][w] < 0) continue;

                        DTNHost hi = getHost();                        
                        ri2.addMoreInfo(new RoutingInfo(String.format("(%d,%d) -> %s: %.3f", hi.getAddress(), hj.getKey().getAddress(), w, routeDelays[j][w])));      
                    }

		top.addMoreInfo(ri2);
		return top;            
	}

	@Override
	public MessageRouter replicate() {
		ContactRouter r = new ContactRouter(this);
		return r;
	}

	private List<String> getDeliveredMessages()
	{
		return deliveredMessages;
	}

	public boolean containsKey(Map<PeerEvent, PeerEventEntry> map, PeerEvent key)
	{
		for (Map.Entry<PeerEvent, PeerEventEntry> e : map.entrySet())
			if (key.equals(e.getKey()))
				return true;
		return false;
	}

	public PeerEventEntry getValue(Map<PeerEvent, PeerEventEntry> map, PeerEvent key)
	{
		for (Map.Entry<PeerEvent, PeerEventEntry> e : map.entrySet())
			if (key.equals(e.getKey()))
				return e.getValue();
		return null;
	}

	public RoutingTable getValue(Map<Contact, RoutingTable> map, Contact key)
	{
		for (Map.Entry<Contact, RoutingTable> e : map.entrySet())
			if (key.equals(e.getKey()))
				return e.getValue();
		return null;
	}	

	private double getDelay_ij_ik(DTNHost hj, DTNHost hk)
	{
		int j, k;
		double factor;

		j = getNeighborIndex(hj);
		k = getNeighborIndex(hk);

		factor = Math.exp(Math.pow(5.0/neighbors.get(hk).getVisitedCounter(), 2));
		//if (contactCounts[j][k] < 5 || contactQs[j][k] < 0.90) 
			//    return -1.0;

		if (contactCounts[j][k] > 3)
			//return (contactQs[j][k] * contactDelays[j][k] + (1 - contactQs[j][k]) * DELAY_MAX);                
		//return Math.exp(Math.pow((1 - contactQs[j][k])/ (1 - gamma), 2.0)) * (1 + eta) * contactDelays[j][k];
                    return 1.0 * (contactDelays[j][k]  + 1.28 * Math.sqrt(contactVariance[j][k] /contactCounts[j][k]));
		else 
                    return -1.0;
	}


	private double getDelay_ij_w(DTNHost hj, DTNHost hw)
	{
		int j, w;

		if (hj.equals(hw)) return 0.0;

		j = getNeighborIndex(hj);
		w = getDestinationIndex(hw);

		return routeDelays[j][w];
	}

	private double getDelay_ij_w(DTNHost hj, int w)
	{
		int j;
                
		if (hj.getAddress() == w ) 
		{
                        if (keepContactList)
                        {
                            List<DTNHost> list = contactList.get(hj+"_"+w);
                            if (list == null)
                            {
                                    contactList.put(hj + "_" + w, new ArrayList<DTNHost>());
                                    list = contactList.get(hj + "_" + w);                
                                    list.clear();
                                    list.add(hj); // hj == w
                            }
                        }
			return 0.0;
		}

		j = getNeighborIndex(hj);            

		return routeDelays[j][w];
	}

	private void setDelay_ij_ik(DTNHost hj, DTNHost hk, double delay)
	{
		int j, k, n;
		double f;

		j = getNeighborIndex(hj);
		k = getNeighborIndex(hk);
		n = contactCounts[j][k];
	
		//f = 7.0/8.0;

		//log(getHost() + " :: " + hj + "=>" + hk + " = " + delay);
                
                if (n == 0)
                    contactDelays[j][k] = delay;
                else
                    contactDelays[j][k] = (n * contactDelays[j][k] + delay) / (n + 1);
		
                if (n >= 1)
		{
			contactVariance[j][k] = ((n - 1) * contactVariance[j][k] + (contactDelays[j][k] - delay) * (contactDelays[j][k] - delay)) / n;
			//contactVariance[j][k] = 0.75 * contactVariance[j][k] / n + 0.25 * delay * delay;
			//contactQs[j][k] = 2 * Phi(eta * contactDelays[j][k] * Math.sqrt(contactDelays[j][k]/contactVariance[j][k])) - 1.0;
			contactQs[j][k] = Phi(eta * contactDelays[j][k] * Math.sqrt(contactDelays[j][k]/contactVariance[j][k]));
			//if (contactQs[j][k] < 0 || contactQs[j][k] > 1) 
			//    contactQs[j][k] = 0.0;
		}
		else
			contactQs[j][k] = 0.0;

		//contactDelays[j][k] = contactQs[j][k] * contactDelays[j][k] + (1 - contactQs[j][k]) * DELAY_MAX;
		//contactDelays[j][k] = contactDelays[j][k] + 4 * contactVariance[j][k];
		contactCounts[j][k] = n + 1;
	}

	private void setDelay_ij_w(DTNHost hj, int w)
	{
		MessageRouter oRouter = hj.getRouter();
		assert oRouter instanceof ContactRouter : "contact only works " + 
		" with other routers of same type";

		if (w == getHost().getAddress()) return;

		ContactRouter otherRouter = (ContactRouter)oRouter;

		/* 
                    c(ij --> w) = c(ji --> jl) + c(jl --> w)
                    c(ij --> j) = 0 
		*/

		int j;            

		j = getNeighborIndex(hj);            

		if (hj.getAddress() == w)
		{
			routeDelays[j][w] = 0.0;
                        if (keepContactList)
                        {
                            List<DTNHost> list = contactList.get(hj+"_"+w);
                            if (list == null)
                            {
                                    contactList.put(hj + "_" + w, new ArrayList<DTNHost>());
                                    list = contactList.get(hj + "_" + w);
                            }
                            list.clear();
                            list.add(hj); // hj == w
                        }
			return;
		}            

		Map<DTNHost, NeighborEntry> otherNeighbors = otherRouter.getNeighbors();                           
		double temp = -1.0;
		int tempHop = 0;
		DTNHost candNextHop = null;

		for(Map.Entry<DTNHost, NeighborEntry> e : otherNeighbors.entrySet())
		{
			DTNHost hl = e.getKey();
			if (hl.equals(this.getHost())) continue;                               

			double delay_ji_jl = otherRouter.getDelay_ij_ik(getHost(), hl);
			if (delay_ji_jl < 0) continue;

			double delay_jl_w = otherRouter.getDelay_ij_w(hl, w);                
			if (delay_jl_w < 0) continue;

			if (temp < 0 || temp > delay_ji_jl + delay_jl_w)
			{
				temp = delay_ji_jl + delay_jl_w;                    
				tempHop = 1 + otherRouter.getRouteHopCount(hl, w);
				candNextHop = hl;
			}
		}

                if (temp > 0)
                {
                    //log(getHost() + " :: " + hj + "===>" + w + " = " + temp);
                    routeDelays[j][w] = temp;
                    routeHopCount[j][w] = tempHop;
                }
                
		if (keepContactList)
                    if (temp >= 0.0 && otherRouter.contactList.get(candNextHop+"_"+w) != null){				

                            List<DTNHost> list = contactList.get(hj+"_"+w);
                            List<DTNHost> otherList = otherRouter.contactList.get(candNextHop+"_"+w);

                            if (list == null)
                            {
                                    contactList.put(hj+"_"+w, new ArrayList<DTNHost>());
                                    list = contactList.get(hj+"_"+w);
                            }                 
                            list.clear();                
                            list.add(hj);
                            if (otherList.contains(hj))
                            {
                                    int index = otherList.indexOf(hj);
                                    while(++index < otherList.size())
                                            list.add(otherList.get(index));
                            }
                            else                    
                            {                    
                                    for(DTNHost h : otherList)
                                            list.add(h);                    
                            }                                
                    }
	}

	private int getRouteHopCount(DTNHost hh, int w)
	{
		int h = getNeighborIndex(hh);
		return routeHopCount[h][w];        
	}

	private Map<DTNHost, NeighborEntry> getNeighbors()
	{
		return neighbors;
	}

	private int getDestinationIndex(DTNHost h)
	{
		return h.getAddress();
	}

	private int getNeighborIndex(DTNHost h)
	{
		int index = 0;
		for (Map.Entry<DTNHost, NeighborEntry> e : neighbors.entrySet())
		{
			if (h.equals(e.getKey())) return index;
			index++;
		}
		return 0;
	}

	private double Phi(double x)
	{
		final double b1 =  0.319381530;
		final double b2 = -0.356563782;
		final double b3 =  1.781477937;
		final double b4 = -1.821255978;
		final double b5 =  1.330274429;
		final double p  =  0.2316419;
		final double c  =  0.39894228;

		if(x >= 0.0) {
			double t = 1.0 / ( 1.0 + p * x );
			return (1.0 - c * Math.exp( -x * x / 2.0 ) * t *
					( t *( t * ( t * ( t * b5 + b4 ) + b3 ) + b2 ) + b1 ));
		}
		else {
			double t = 1.0 / ( 1.0 - p * x );
			return ( c * Math.exp( -x * x / 2.0 ) * t *
					( t *( t * ( t * ( t * b5 + b4 ) + b3 ) + b2 ) + b1 ));
		}
	}  

	private static void log(String msg)
	{
		if (debug)
			System.out.println(SimClock.getTime() + " " + msg);
	}

	private class ContactMessage extends Message{
		public boolean forwarded;        
		public List<NextNode> toForward;        
		public  List<DTNHost> list;
		public int nextHopIndex;
		public int copiesLeft;
		public boolean isOnRoute;

		public double lastForwarded;
		public double estimatedDelay;


		public ContactMessage(DTNHost from, DTNHost to, String id, int size) {
			super(from, to, id, size, msgTtl);
			forwarded = false;            
			toForward = null;
			list = null;
			nextHopIndex = 0;
			copiesLeft = messageCopies;
			isOnRoute = false;
		}

		@Override
		public ContactMessage replicate() {
			ContactMessage m = new ContactMessage(getFrom(),getTo(),getId(),getSize());
			m.startTime = startTime;
			m.forwarded = false;            
			m.toForward = null;
			m.nextHopIndex = nextHopIndex;
			m.copiesLeft = copiesLeft;
			m.isOnRoute = isOnRoute;

			m.estimatedDelay = this.estimatedDelay;
			m.lastForwarded = this.lastForwarded;

			if (list != null)
				m.list = new ArrayList<DTNHost>(list);
			m.copyFrom(this);                
			return m;
		}

		@Override
		public String toString() {
			String s="";
			if (list != null)
				for(DTNHost h: list)
					s += h + " ";
			return getId()+ "(" + getFrom() + "=>" + getTo() + ") ["+s+"] ";
		}

		public boolean isForwarded()                
		{
			return forwarded;
		}

		public void setToForward(List<NextNode> h)
		{
			toForward = h;        
		}

		public List<NextNode> getToForward()
		{
			return toForward;
		}

		public void setTravelList(List<DTNHost> l)
		{
			this.list = l;
			nextHopIndex = 0;
		}       

		public DTNHost getNextHop()
		{
			if (list == null) return null;
			if(nextHopIndex < list.size())
				return list.get(nextHopIndex);
			return null;
		}

		public List<DTNHost> getList()
		{
			return list;
		}

		public void advanceNextHopIndex()
		{
			nextHopIndex++;
		}
	}        
}

