/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package routing;

/**
 *
 * @author Yusuf Sarwar
 */
/* 
 * Copyright 2007 TKK/Netlab
 * Released under GPLv3. See LICENSE.txt for details. 
 */

/* Hop count metric */

import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import core.Connection;
import core.DTNHost;
import core.Message;
import core.Settings;
import core.SimClock;
import core.Tuple;

import java.util.Collections;
import routing.contact.*;
/**
 * Implementation of contact router as described in 
 * <I>Probabilistic routing in intermittently connected networks</I> by
 * Anders Lindgren et al.
 */
public class MetricContactRouter extends ActiveRouter {
		/** contact router's setting namespace ({@value})*/ 
	public static final String contact_NS = "ContactRouter";
        
    private static boolean debug;
        /** Contact lists */
        private Map<PeerEvent, PeerEventEntry> contacts;    
        private Map<DTNHost, NeighborEntry> visitedNodes;
        private Map<DTNHost, NeighborEntry> neighbors;        
        
        private int noOfNeighbors, noOfTotalNodes;
        private double contactDelays[][], contactSD2[][], contactQs[][], routeDelays[][];
        private int routeHopCount[][];
        private DTNHost nextHop[][];
        private int contactCounts[][];
        private Map<Message, Double> waitProps;        
        
        private Map<String, List<DTNHost>> contactList;
        private List<String> forwardedMessages;
        private final int DELAY_MAX = 2000;
        private final double eta = 0.1;
        private boolean newConnection = false, toBeForwarded = false;
        
        private List<String> deliveredMessages;
        
     /**
	 * Constructor. Creates a new message router based on the settings in
	 * the given Settings object.
	 * @param s The settings object
	 */
	public MetricContactRouter(Settings s) {
		super(s);        
        
        s.setNameSpace(contact_NS);		
        noOfTotalNodes = s.getInt("nrofHosts");
        noOfNeighbors = noOfTotalNodes/5;
        
		if (s.contains("debug")) {
			debug = s.getBoolean("debug");
		}
		else {
			debug = true;
		}

		initContacts();
	}

	/**
	 * Copyconstructor.
	 * @param r The router prototype where setting values are copied from
	 */
	protected MetricContactRouter(MetricContactRouter r) {
            super(r);      
            noOfTotalNodes = r.noOfTotalNodes;
            noOfNeighbors = r.noOfNeighbors;
            initContacts();
	}
	
	/**
	 * Initializes predictability hash
	 */
	private void initContacts() {           
        
        //this.contacts = new HashMap<PeerEvent, PeerEventEntry>();
        this.visitedNodes = new HashMap<DTNHost, NeighborEntry>();
        this.neighbors = new HashMap<DTNHost, NeighborEntry>();
        this.forwardedMessages = new ArrayList<String>();
        //this.routingTables = new HashMap<Contact, RoutingTable>();            
        //forwardingEvents =  new HashMap<Message, DTNHost>();            

        this.contactDelays = new double[noOfNeighbors][noOfNeighbors];
        this.contactSD2 = new double[noOfNeighbors][noOfNeighbors];
        this.contactQs = new double[noOfNeighbors][noOfNeighbors];
        this.contactCounts = new int[noOfNeighbors][noOfNeighbors];
        this.waitProps = new HashMap<Message, Double>();
        
        this.contactList = new HashMap<String, List<DTNHost>>();
            
        this.routeDelays = new double[noOfNeighbors][noOfTotalNodes];
        this.routeHopCount = new int[noOfNeighbors][noOfTotalNodes];
        this.nextHop = new DTNHost[noOfNeighbors][noOfTotalNodes];

        for(int i = 0; i < noOfNeighbors; i++)
            for (int j = 0; j < noOfTotalNodes; j++)
            {
                routeDelays[i][j] = -1.0;                      
                routeHopCount[i][j] = 0;                                
            }
        deliveredMessages = new ArrayList<String>();                
	}

	@Override
	public void changedConnection(Connection con) {
            if (con.isUp()) {                
                DTNHost otherHost = con.getOtherNode(getHost());                
                meetANode(otherHost);
                if (visitedNodes.get(otherHost).getVisitedCounter() > 10)
                {                    
                    updateContactFor(otherHost);
                    
                    /*if (contactTimes != null && SimClock.getTime() > day * 1000)
                    {
                        day++;
                        System.out.println("Day " + day);
                        for(Map.Entry<String, List<Double>> e : contactTimes.entrySet() )
                        {
                            String key = e.getKey();
                            List<Double> l = e.getValue();
                            System.out.println(key + " " + l);
                        }
                    }
                    */
                    updateContacts(otherHost);                                        
                    newConnection = true;
                }
            }
	}
	
    private void meetANode(DTNHost host)
    {
        NeighborEntry ne = visitedNodes.get(host);
        if (ne == null)
        {
            visitedNodes.put(host, new NeighborEntry(neighbors.size()));
            ne = visitedNodes.get(host);
        }            
            
        ne.setVisitedCounter(ne.getVisitedCounter() + 1);
        ne.setLastVisited(SimClock.getTime());
        
        
        MetricContactRouter otherRouter = (MetricContactRouter)host.getRouter();
        List<String> list =  otherRouter.getDeliveredMessages();
        
        if (list.size() > 0)
            for (String m : list) {
                if (!deliveredMessages.contains(m))            
                    deliveredMessages.add(m);               
                if (this.getMessage(m) != null)
                    deleteMessage(m, false);
            }
    }
    
	/**
	 * Updates Contact for a host.
	 * <CODE>T(ab, ac) = alpha * T(ab, ac)_old + (1 - alpha) * (ab, ac) </CODE>
	 * @param host The host we just met
	 */
	private void updateContactFor(DTNHost host) 
        {
            if (!neighbors.containsKey(host) && neighbors.size() >= noOfNeighbors) return;
            
            /* Find pair-wise meeting contact with this host */
            for (Map.Entry<DTNHost, NeighborEntry> e : neighbors.entrySet())                
            {     
                /* ij -> ik happens i: this node, j: neighbor, k: just visited */
                if (e.getKey().equals(host)) continue; /* not the current node it meets */
                double lastContactTime = SimClock.getTime() - e.getValue().getLastVisited(); 
                
                setDelay_ij_ik(e.getKey(), host, lastContactTime);
                
                /* contact delay */
                /*
                if (getHost().getAddress() == 11 && contactTimes == null)
                    contactTimes = new HashMap<String, List<Double>>();
                
                if (contactTimes != null)
                {
                    String key = e.getKey()+"_"+host;
                    List<Double> l = contactTimes.get(key);
                    if (l == null)
                    {
                        contactTimes.put(key, new ArrayList<Double>());
                        l = contactTimes.get(key);
                    }
                    l.add(lastContactTime);                    
                }
                 */
                /* contact delay */
            }            
            
            NeighborEntry nentry = neighbors.get(host);
            if (nentry != null)
            {
                nentry.setLastVisited(SimClock.getTime());
                nentry.setVisitedCounter(nentry.getVisitedCounter() + 1);
            }
            else
                if (neighbors.size() < noOfNeighbors) 
                    neighbors.put(host, new NeighborEntry(SimClock.getTime(), visitedNodes.get(host).getVisitedCounter(), neighbors.size()));
            
            /* Penalty for not meeting for a while */
            List<DTNHost> dels = new ArrayList<DTNHost>();
            
            for(Map.Entry<DTNHost, NeighborEntry> e : neighbors.entrySet())
            {
                DTNHost h1 = e.getKey();
                double mt1 = e.getValue().getLastVisited();
                
                /*
                if (SimClock.getTime() - mt1 > DELAY_MAX && SimClock.getTime() - e.getValue().getPenaltyTime() > DELAY_MAX)
                {
                    // h1 hasn't been visited more than 5000 time    
                    log(getHost() + " Penalty for " + h1 + " " + (SimClock.getTime() - mt1));
                    for(Map.Entry<DTNHost, NeighborEntry> eo : neighbors.entrySet())
                    {
                        DTNHost h2 = eo.getKey();
                        if (h2.equals(h1)) continue;
                        int j, k;
                        j = getNeighborIndex(h2);
                        k = getNeighborIndex(h1);
                        contactQs[j][k] = 0.6 * contactQs[j][k];
                    }                    
                    e.getValue().setPenaltyTime(SimClock.getTime());
                }
                */
                
                if (SimClock.getTime() - mt1 > DELAY_MAX)
                {
                    int counter = neighbors.get(h1).getVisitedCounter();
                    counter--;
                    if (counter < 3)
                    {
                        dels.add(h1);
                    }
                    else
                    {
                        neighbors.get(h1).setVisitedCounter(counter);
                        neighbors.get(h1).setLastVisited(SimClock.getTime());
                    }
                }
            }            
        }     
      	
	/**
	 * Returns the current prediction (P) value for a host or 0 if entry for
	 * the host doesn't exist.
	 * @param host The host to look the P for
	 * @return the current P value
	 */
	
	
	/**
	 * Updates transitive (A->B->C) delivery predictions.
	 * <CODE>P(a,c) = P(a,c)_old + (1 - P(a,c)_old) * P(a,b) * P(b,c) * BETA
	 * </CODE>
	 * @param host The B host who we just met
	 */
	private void updateContacts(DTNHost host) {                                  
            
            if (!neighbors.containsKey(host)) return;            
            DTNHost hj = host;
            /* for each destination, dest w, compute c(ij-->w) */
            int dest;
            for(dest = 0; dest < noOfTotalNodes; dest++)
            {
                if (dest == getHost().getAddress()) continue;
                setDelay_ij_w(hj, dest);            
            }        
	}

        void showContacts(Map<PeerEvent, PeerEventEntry> map, String name)
        {
            log(name);
            for (Map.Entry<PeerEvent, PeerEventEntry> e : map.entrySet())
                log(e.getKey() + " : " + e.getValue());
        }       
               
        /**
         * Schedules messages originated from this very node to some destination 
         */
                
	@Override
	public void update() {            
            super.update();
            if (!canStartTransfer() || isTransferring()) {
                    return; // nothing to transfer or is currently transferring 
            }

            // try messages that could be delivered to final recipient
            if (exchangeDeliverableMessages() != null) {
                    return;
            }

            tryOtherMessages();		
	}
	
    @Override
	public int receiveMessage(Message m, DTNHost from) {		            
            int ret = super.receiveMessage(m, from);
            //log("\t\t" + getHost() + " treats " + m +  " from " + from + " as " + ret);
            return ret;
	}
        
    @Override 
	public boolean createNewMessage(DTNHost to, String id, int size,
			int responseSize) {		
        
            makeRoomForNewMessage(size);		
            ContactMessage m = new ContactMessage(getHost(), to, id, size);
            m.setResponseSize(responseSize);
            addToMessages(m, true);				
            return true;
	}
        
	/**
	 * Tries to send all other messages to all connected hosts ordered by
	 * their delivery probability
	 * @return The return value of {@link #tryMessagesForConnected(List)}
	 */
	
         private Tuple<Message, Connection> tryOtherMessages() {            
            List<Tuple<Message, Connection>> messages =                     
                    new ArrayList<Tuple<Message, Connection>>(); 
            
            //Collection<Message> msgCollection = getMessageCollection();
                        
            for (Connection con : getConnections()) 
            {
                DTNHost other = con.getOtherNode(getHost());
                
                //if (neighbors.containsKey(other) == false) continue; /* Not a neighbor node */
                
                MetricContactRouter otherRouter = (MetricContactRouter)other.getRouter();                

                if (otherRouter.isTransferring()) {
                        continue; // skip hosts that are transferring
                }                              
                
                for (Message m : getMessageCollection())
                {
                    assert m instanceof ContactMessage;                   
                   
                    // Check if the message is there or visited                    
                    if (m.getHops().contains(other) || otherRouter.getMessage(m.getId()) != null) continue; // the message already there
                                        
                    // Deciding the next hop for m
                    ContactMessage cm = (ContactMessage)m;                    
                                        
                    if (cm.forwarded) continue; // already forwarded
                    
                    if (cm.getToForward() != null && cm.getToForward().equals(other)) // scheduled before and now in meet
                    {
                        // already scheduled to this node
                        log(getHost() + " forwards " + m + " to " + other + " prescheduled " + SimClock.getTime());
                        //cm.forwarded = true;
                        messages.add(new Tuple<Message, Connection>(m,con));
                        continue;
                    }
                    
                    DTNHost dest = m.getTo();                    
                    
                    double delay_ij_w = getDelay_ij_w(other, dest);

                    //log("Delay  " + delay_ij_w);
                    
                    if (delay_ij_w < 0) 
                    {
                        // No routing table entry... forward and keep copy
                        // messages.add(new Tuple<Message, Connection>(m, con));
                        // if (forwardedMessages.contains(m.getId()) == false)
                        //    forwardedMessages.add(m.getId());
                        /*
                        if (cm.copiesLeft > 1)
                        {
                            log(getHost() + " gives " + m + " " + cm.copiesLeft + " to " + other);
                            messages.add(new Tuple<Message, Connection>(m, con));                            
                        }
                        */
                        
                        continue;
                        
                        // May be the node j is not the neighbor of current node i                        
                        //delay_ij_w = otherRouter.getBestRouteToDestination(dest);
                        /*
                        if (delay_ij_w < 0)
                        {
                            // Not routing path exists... do flooding
                            log(this.getHost() + " floods " + m.getId() + " to " + other);
                            messages.add(new Tuple<Message, Connection>(m, con));
                            continue;                        
                        }
                        */
                    }
                                                            
                    double minDelay; 
                    int hop = 0;
                    DTNHost candHost;                    
                    
                    minDelay = delay_ij_w;
                    candHost = other;
                                        
                    for (Map.Entry<DTNHost, NeighborEntry> e : neighbors.entrySet())
                    {
                        if (e.getKey().equals(other) || m.getHops().contains(e.getKey())) continue;
                        
                        DTNHost hk = e.getKey();                        
                        double delay_ij_ik = getDelay_ij_ik(other, hk);
                        if (delay_ij_ik < 0) continue;
                        
                        double delay_ik_w = getDelay_ij_w(hk, dest);                                               
                        if (delay_ik_w < 0) continue;
                        
                        double alternateDelay = delay_ij_ik + delay_ik_w;
                        
                        if (alternateDelay < minDelay)
                        {
                            minDelay = alternateDelay;
                            candHost = hk;
                            //hop = routeHopCount[getNeighborIndex(hk)][dest.getAddress()];
                        }
                        //if (minDelay < delay_ij_w)
                        //    break;
                    }

                    //log("MinDelay  " + minDelay);                                        
                    
                    toBeForwarded = false;
                    if (newConnection)
                    {                    
                        if (minDelay < delay_ij_w)
                        {
                            
                        }
                        else                            
                            toBeForwarded = true;
                        newConnection = false;
                    }                    
                                        
                    if (cm.getList() == null)
                    {                        
                        cm.setTravelList(contactList.get(other+"_"+cm.getTo().getAddress()));                        
                    }
                    
                    candHost = cm.getNextHop(); /* Comment this line for non-source routing */
                    
                    if (candHost == null && newConnection){
                        log(getHost() + " fails to forward " + m + " to " + other);
                        newConnection = false;
                        continue; // no suitable candidate is obtained
                    }
                    
                    if (candHost == null) continue;
                    
                    if (candHost.getAddress() == other.getAddress()) {
                        ContactMessage conm = (ContactMessage)m;
                        log(getHost() + " forwards " + m + " to " + other + " " + SimClock.getTime() + " ed " + conm.estimatedDelay); 
                        
                        if (conm.isOnRoute == false)
                        {
                            conm.isOnRoute = true;
                            conm.setStartTime(SimClock.getTime());
                            conm.setTravelList(contactList.get(candHost+"_"+conm.getTo().getAddress()));
                            conm.estimatedDelay = minDelay;
                        } 
                                                
                        messages.add(new Tuple<Message, Connection>(m,con));
                    }                    
                    else
                    {  
                        ContactMessage conm = (ContactMessage)m;
                        if (conm.getToForward() != candHost)
                        {
                            log(getHost() + " schedules " + m + " for " + candHost + " @" + SimClock.getTime());

                            if (conm.getFrom() == getHost())
                                conm.setStartTime( SimClock.getTime());                        
                            conm.setToForward(candHost);                        
                        }
                    }                    
                }
            }
            
            if (messages.size() == 0) {
                    return null;
            }

            // sort the message-connection tuples
            Collections.sort(messages, new TupleComparator());
            return tryMessagesForConnected(messages);	// try to send messages
	}      
        
         public double getBestRouteToDestination(DTNHost dest)
         {
            double leastDelay = Double.MAX_VALUE;
            
            for (Map.Entry<DTNHost, NeighborEntry> e : neighbors.entrySet())
            {
                DTNHost next = e.getKey();                
                double currDelay =  getDelay_ij_w(next, dest);
                
                if (currDelay < leastDelay)
                {
                    leastDelay = currDelay;                    
                }
            }
            return leastDelay;
         }
         
    @Override
	public Message messageTransferred(String id, DTNHost from) {
        
            ContactMessage msg = (ContactMessage)super.messageTransferred(id, from);				
            msg.advanceNextHopIndex();
            if (msg.isOnRoute == false)
                msg.copiesLeft = (int)Math.ceil(msg.copiesLeft/2.0);		
            msg.forwarded = false;
            return msg;
	}  
         
    @Override
	protected void transferDone(Connection con) {            
            Message m = con.getMessage();            
            DTNHost other = con.getOtherNode(getHost());
            
            log(getHost() + " transfer done " + m + " to " + other);
            
            Message msg = this.getMessage(m.getId());
            ContactMessage cmsg = (ContactMessage)msg;
            cmsg.forwarded = true;
            
            cmsg.estimatedDelay = cmsg.estimatedDelay - (SimClock.getTime() - cmsg.lastForwarded);            
            if (cmsg.estimatedDelay < 0)
                cmsg.estimatedDelay = 0.0;
            
            cmsg.lastForwarded = SimClock.getTime();
            
            if (other.equals(m.getTo()))
            {               
                /* Destination reached at the other end */
                String hops="";                        
                for(DTNHost h : m.getHops())
                    hops = hops + " " + h.toString() ;              
                hops = "C: " + m.getCreationTime() + " S: " + ((ContactMessage)m).getStartTime() + " D: " + SimClock.getTime() + " " + hops;
                log(m + " (" + m.getFrom() + " -> " + m.getTo() +") : " + hops);
                
                deliveredMessages.add(m.getId());
                this.deleteMessage(m.getId(), false);
            }   
            else
            {                
                if (cmsg.isOnRoute == true)                    
                    cmsg.advanceNextHopIndex();                
                else
                    cmsg.copiesLeft /= 2;                
            }
	}
        
	 /**
	 * Comparator for Message-Connection-Tuples that orders the tuples by
	 * their delivery probability by the host on the other side of the 
	 * connection (GRTRMax)
	 */
	private class TupleComparator implements Comparator                 
            <Tuple<Message, Connection>> {
            
            public int compare(Tuple<Message, Connection> tuple1,
                            Tuple<Message, Connection> tuple2) {                
                
                ContactMessage m1 = (ContactMessage)tuple1.getKey();
                ContactMessage m2 = (ContactMessage)tuple2.getKey();
                
                return (int)(m1.estimatedDelay - m2.estimatedDelay);
            }            
	}
	
	@Override
	public RoutingInfo getRoutingInfo() {        

            RoutingInfo top = super.getRoutingInfo();
            RoutingInfo ri = new RoutingInfo(neighbors.size() + 
                            " neighors(s)");

            for (Map.Entry<DTNHost, NeighborEntry> e: neighbors.entrySet())
                ri.addMoreInfo(new RoutingInfo(String.format("%s meet %s", e.getKey(), e.getValue())));      
            top.addMoreInfo(ri);
            
            RoutingInfo ri1 = new RoutingInfo(neighbors.size() + 
                            " contacts(s)");

            for(Map.Entry<DTNHost, NeighborEntry> hj : neighbors.entrySet())
            for(Map.Entry<DTNHost, NeighborEntry> hk : neighbors.entrySet())
            {
                int j, k;
                j = getNeighborIndex(hj.getKey());
                k = getNeighborIndex(hk.getKey());
                
                if (j == k || contactCounts[j][k] == 0) continue;
                
                DTNHost hi = getHost();
                        
                ri1.addMoreInfo(new RoutingInfo(String.format("(%s,%s) -> (%s,%s) = %.3f :: %.3f", hi, hj.getKey(), hi, hk.getKey(), contactDelays[j][k], contactQs[j][k])));      
            }
            top.addMoreInfo(ri1);
            
            RoutingInfo ri2 = new RoutingInfo(neighbors.size() + " contacts(s)");

            for(Map.Entry<DTNHost, NeighborEntry> hj : neighbors.entrySet())
            for(int w = 0; w < noOfTotalNodes; w++)
            {
                int j, k;
                
                j = getNeighborIndex(hj.getKey());                                                
                
                if (routeDelays[j][w] < 0) continue;
                
                DTNHost hi = getHost();                        
                ri2.addMoreInfo(new RoutingInfo(String.format("(%d,%d) -> %s: %.3f", hi.getAddress(), hj.getKey().getAddress(), w, routeDelays[j][w])));      
            }
            
            top.addMoreInfo(ri2);
            return top;            
	}
	
	@Override
	public MessageRouter replicate() {
            MetricContactRouter r = new MetricContactRouter(this);
            return r;
	}
        
    private List<String> getDeliveredMessages()
    {
        return deliveredMessages;
    }
    
        public boolean containsKey(Map<PeerEvent, PeerEventEntry> map, PeerEvent key)
        {
            for (Map.Entry<PeerEvent, PeerEventEntry> e : map.entrySet())
                if (key.equals(e.getKey()))
                    return true;
            return false;
        }
        
        public PeerEventEntry getValue(Map<PeerEvent, PeerEventEntry> map, PeerEvent key)
        {
            for (Map.Entry<PeerEvent, PeerEventEntry> e : map.entrySet())
                if (key.equals(e.getKey()))
                    return e.getValue();
            return null;
        }
        
        public RoutingTable getValue(Map<Contact, RoutingTable> map, Contact key)
        {
            for (Map.Entry<Contact, RoutingTable> e : map.entrySet())
                if (key.equals(e.getKey()))
                    return e.getValue();
            return null;
        }
        
        private double getDelay_ij_ik(DTNHost hj, DTNHost hk)
        {
            int j, k;
            
            j = getNeighborIndex(hj);
            k = getNeighborIndex(hk);
            
            //if (contactCounts[j][k] < 5 || contactQs[j][k] < 0.90) 
            //    return -1.0;
                
            if (contactCounts[j][k] > 5)
                //return (contactQs[j][k] * contactDelays[j][k] + (1 - contactQs[j][k]) * DELAY_MAX);
                return contactDelays[j][k];
            else
                return -1.0;
        }
        
        
        private double getDelay_ij_w(DTNHost hj, DTNHost hw)
        {
            int j, w;
            
            if (hj.equals(hw)) return 0.0;
            
            j = getNeighborIndex(hj);
            w = getDestinationIndex(hw);
            
            return routeDelays[j][w];
        }
        
        private double getDelay_ij_w(DTNHost hj, int w)
        {
            int j;
            
            if (hj.getAddress() == w) 
            {
                List<DTNHost> list = contactList.get(hj+"_"+w);
                if (list == null)
                {
                    contactList.put(hj + "_" + w, new ArrayList<DTNHost>());
                    list = contactList.get(hj + "_" + w);                
                    list.clear();
                    list.add(hj); // hj == w
                }
                 return 0.0;
            }
            
            j = getNeighborIndex(hj);            
            
            return routeDelays[j][w];
        }
        
        /* c(ij->ik) = e ^ (2 * (1 - f_i^k))^2 */
        
        private void setDelay_ij_ik(DTNHost hj, DTNHost hk, double delay)
        {
            int j, k, n;
            double f = 0.0;
            
            j = getNeighborIndex(hj);
            k = getNeighborIndex(hk);
            
            if (neighbors.get(hk) == null)
                return;
            
            int totalCount, thisCount;
            
            totalCount = 0;
            thisCount = 0;
            for(Map.Entry<DTNHost, NeighborEntry> e : neighbors.entrySet())
            {
                if (e.equals(hk))
                    thisCount = e.getValue().getVisitedCounter();
                totalCount += e.getValue().getVisitedCounter();                
            }
            f = (double)(1.0 * thisCount) / totalCount;
            
            for(j=0; j < noOfNeighbors; j++)
            {
                if (j == k) continue;
                //contactDelays[j][k] = Math.pow(Math.E, Math.pow(20.0 / thisCount, 2)); // threshold
                contactDelays[j][k] = Math.pow(Math.E, Math.pow(2 * (1 - f), 2)); // metric
                contactCounts[j][k] = contactCounts[j][k] + 1;
            }
        }
        
        private void setDelay_ij_w(DTNHost hj, int w)
        {
            MessageRouter oRouter = hj.getRouter();
            assert oRouter instanceof MetricContactRouter : "contact only works " + 
                    " with other routers of same type";
                        
            if (w == getHost().getAddress()) return;
            
            MetricContactRouter otherRouter = (MetricContactRouter)oRouter;
            
            /* 
             c(ij --> w) = c(ji --> jl) + c(jl --> w)
             c(ij --> j) = 0 
            */
            
            int j, l;            
            
            j = getNeighborIndex(hj);            
            
            if (hj.getAddress() == w)
            {
                routeDelays[j][w] = 0.0;
                List<DTNHost> list = contactList.get(hj+"_"+w);
                if (list == null)
                {
                    contactList.put(hj + "_" + w, new ArrayList<DTNHost>());
                    list = contactList.get(hj + "_" + w);
                }
                list.clear();
                list.add(hj); // hj == w
                return;
            }            
            
            Map<DTNHost, NeighborEntry> otherNeighbors = otherRouter.getNeighbors();                           
            double temp = -1.0;
            int tempHop = 0;
            DTNHost candNextHop = null;
            
            for(Map.Entry<DTNHost, NeighborEntry> e : otherNeighbors.entrySet())
            {
                DTNHost hl = e.getKey();
                if (hl.equals(this.getHost())) continue;                               
                
                double delay_ji_jl = otherRouter.getDelay_ij_ik(hj, hl);
                if (delay_ji_jl < 0) continue;
                
                double delay_jl_w = otherRouter.getDelay_ij_w(hl, w);                
                if (delay_jl_w < 0) continue;
                
                if (temp < 0 || temp > delay_ji_jl + delay_jl_w)
                {
                    temp = delay_ji_jl + delay_jl_w;                    
                    tempHop = 1 + otherRouter.getRouteHopCount(hl, w);
                    candNextHop = hl;
                }
            }
            
            if (temp >= 0.0 && otherRouter.contactList.get(candNextHop+"_"+w) != null){
                routeDelays[j][w] = temp;
                routeHopCount[j][w] = tempHop;
                
                List<DTNHost> list = contactList.get(hj+"_"+w);
                List<DTNHost> otherList = otherRouter.contactList.get(candNextHop+"_"+w);
                
                if (list == null)
                {
                    contactList.put(hj+"_"+w, new ArrayList<DTNHost>());
                    list = contactList.get(hj+"_"+w);
                }                 
                list.clear();                
                list.add(hj);
                if (otherList.contains(hj))
                {
                    int index = otherList.indexOf(hj);
                    while(++index < otherList.size())
                        list.add(otherList.get(index));
                }
                else                    
                {                    
                    for(DTNHost h : otherList)
                        list.add(h);                    
                }                                
            }
        }
        
        private int getRouteHopCount(DTNHost hh, int w)
        {
            int h = getNeighborIndex(hh);
            return routeHopCount[h][w];        
        }
        
        private Map<DTNHost, NeighborEntry> getNeighbors()
        {
            return neighbors;
        }
        
        private int getDestinationIndex(DTNHost h)
        {
            return h.getAddress();
        }
        
        private int getNeighborIndex(DTNHost h)
        {
            int index = 0;
            for (Map.Entry<DTNHost, NeighborEntry> e : neighbors.entrySet())
            {
                if (h.equals(e.getKey())) return index;
                index++;
            }
            return 0;
        }
        
        private double Phi(double x)
        {
          final double b1 =  0.319381530;
          final double b2 = -0.356563782;
          final double b3 =  1.781477937;
          final double b4 = -1.821255978;
          final double b5 =  1.330274429;
          final double p  =  0.2316419;
          final double c  =  0.39894228;

          if(x >= 0.0) {
              double t = 1.0 / ( 1.0 + p * x );
              return (1.0 - c * Math.exp( -x * x / 2.0 ) * t *
              ( t *( t * ( t * ( t * b5 + b4 ) + b3 ) + b2 ) + b1 ));
          }
          else {
              double t = 1.0 / ( 1.0 - p * x );
              return ( c * Math.exp( -x * x / 2.0 ) * t *
              ( t *( t * ( t * ( t * b5 + b4 ) + b3 ) + b2 ) + b1 ));
            }
        }  
        
        private static void log(String msg)
        {
            if (debug)
                System.out.println(SimClock.getTime() + " " + msg);
        }

    private class ContactMessage extends Message{
        private boolean forwarded;        
        private DTNHost toForward;        
        private List<DTNHost> list;
        private int nextHopIndex;
        private int copiesLeft;
        private boolean isOnRoute;
        
        public double lastForwarded;
        public double estimatedDelay;
        
        
        public ContactMessage(DTNHost from, DTNHost to, String id, int size) {
            super(from, to, id, size, msgTtl);
            forwarded = false;            
            toForward = null;
            list = null;
            nextHopIndex = 0;
            copiesLeft = 5;
            isOnRoute = false;
        }

        @Override
        public ContactMessage replicate() {
            ContactMessage m = new ContactMessage(getFrom(),getTo(),getId(),getSize());
            m.startTime = startTime;
            m.forwarded = false;            
            m.toForward = null;
            m.nextHopIndex = nextHopIndex;
            m.copiesLeft = copiesLeft;
            m.isOnRoute = isOnRoute;

            m.estimatedDelay = this.estimatedDelay;
            m.lastForwarded = this.lastForwarded;
            
            if (list != null)
                m.list = new ArrayList<DTNHost>(list);
            m.copyFrom(this);                
            return m;
        }
        
        @Override
        public String toString() {
            String s="";
            if (list != null)
            for(DTNHost h: list)
                s += h + " ";
            return getId()+ "(" + getFrom() + "=>" + getTo() + ") ["+s+"] ";
        }

        public boolean isForwarded()                
        {
            return forwarded;
        }
        
        public void setToForward(DTNHost h)
        {
            toForward = h;        
        }
        
        public DTNHost getToForward()
        {
            return toForward;
        }
        
        public void setTravelList(List<DTNHost> l)
        {
            this.list = l;
            nextHopIndex = 0;
        }       
        
        public DTNHost getNextHop()
        {
            if (list == null) return null;
            if(nextHopIndex < list.size())
                return list.get(nextHopIndex);
            return null;
        }

        public List<DTNHost> getList()
        {
            return list;
        }
        
        public void advanceNextHopIndex()
        {
            nextHopIndex++;
        }
    }        
}

