/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package routing;

/**
 *
 * @author Yusuf Sarwar
 */
/* 
 * Copyright 2007 TKK/Netlab
 * Released under GPLv3. See LICENSE.txt for details. 
 */

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import routing.contact.PeerEvent;
import routing.contact.PeerEventEntry;
import core.Connection;
import core.DTNHost;
import core.Message;
import core.Settings;
import core.SimClock;
import core.Tuple;
/**
 * Implementation of contact router as described in 
 * <I>Probabilistic routing in intermittently connected networks</I> by
 * Anders Lindgren et al.
 */

class RouteInfo{
	double delay;
	double variance;
	byte hopCount;        
        List<DTNHost> hops;        
        
	public RouteInfo(double delay, double variance, byte hopCount, List<DTNHost>list) {
		this.delay = delay;
		this.variance = variance;
                this.hopCount = hopCount;
                hops = list;
	}
	public RouteInfo() {
		this.delay = -1.0;
		this.variance = -1.0;
                this.hopCount = 0;                
	}
}

// Stores intercontact time ij -> ik

class InterContactInfo {
	private final static int visitCountThreshold = 3;
	
	double delay;
	byte count;
	double variance;
	double x2;
	
	double last_time_updated;
	
        private static final double lambda = 1.0;
        private static final double gammathr = 0.75;
        private static final double tau = 15 * 60.0; // 15 minutes
        
	public InterContactInfo() {
		this.delay = 0;
		this.count = 0;
		this.variance = 0;
		this.x2 = 0;
		this.last_time_updated = 0.0;
	}
	
	public boolean isValid(double time) {                
		if (time - last_time_updated > (3 * Math.sqrt(variance) + delay))
			return false;
		
                if (count < visitCountThreshold)
			return false;
                
                if (2 * Phi(tau * Math.sqrt(count / variance)) - 1 >= gammathr)
                {
                    //System.out.println("Returing true.");
                    return true;
                }                    
		
		return false;
	}
	
	public void addSample(double last_j_visit_time, double current_time) {
		if (last_j_visit_time <= last_time_updated) 
			return;
		
		double time = current_time - last_j_visit_time;
		delay = (count*delay + time) / (count+1);
		x2 = (count*x2 + time*time) / (count+1);

		if (count > 0) 
			variance = (x2 - delay*delay);

		last_time_updated = current_time;
		count++;		
	}

    private static double Phi(double x)
    {
            final double b1 =  0.319381530;
            final double b2 = -0.356563782;
            final double b3 =  1.781477937;
            final double b4 = -1.821255978;
            final double b5 =  1.330274429;
            final double p  =  0.2316419;
            final double c  =  0.39894228;

            if(x >= 0.0) {
                    double t = 1.0 / ( 1.0 + p * x );
                    return (1.0 - c * Math.exp( -x * x / 2.0 ) * t *
                                    ( t *( t * ( t * ( t * b5 + b4 ) + b3 ) + b2 ) + b1 ));
            }
            else {
                    double t = 1.0 / ( 1.0 - p * x );
                    return ( c * Math.exp( -x * x / 2.0 ) * t *
                                    ( t *( t * ( t * ( t * b5 + b4 ) + b3 ) + b2 ) + b1 ));
            }
    }  

}
public class MulticopyContactRouter extends ActiveRouter {
	/** contact router's setting namespace ({@value})*/ 
	public static final String contact_NS = "ContactRouter";

	private static boolean debug;
	private static final double DEADLINE = 8; // 8 hours
        private static double deadline;
	private int noOfTotalNodes;
	
	private double lastTimeVisited[];
	private InterContactInfo intercontactInfo[][];
	private RouteInfo routeInfo[][];
	
	private List<Message> ACKedMessages;
	private int messageCopies;
	
	List<Tuple<Message, Connection>> forwardList;

	/**
	 * Constructor. Creates a new message router based on the settings in
	 * the given Settings object.
	 * @param s The settings object
	 */
	public MulticopyContactRouter(Settings s) {
		super(s);        

		s.setNameSpace(contact_NS);		
		noOfTotalNodes = s.getInt("nrofHosts");
		
		if (s.contains("debug")) 
			debug = s.getBoolean("debug");
		else 
			debug = true;
		if (s.contains("Deadline"))
                    deadline = s.getDouble("Deadline") * 3600.0;
                else
                    deadline = DEADLINE;
		
                this.messageCopies = s.getInt("messageCopies");
		initContacts();
	}

	/**
	 * Copyconstructor.
	 * @param r The router prototype where setting values are copied from
	 */
	protected MulticopyContactRouter(MulticopyContactRouter r) {
		super(r);      
		noOfTotalNodes = r.noOfTotalNodes;
		messageCopies = r.messageCopies;
		initContacts();
	}
	
	/**
	 * Initializes predictability hash
	 */
	private void initContacts() {           
		this.ACKedMessages = new ArrayList<Message>();

		this.intercontactInfo = new InterContactInfo[noOfTotalNodes][noOfTotalNodes];
		this.routeInfo = new RouteInfo[noOfTotalNodes][noOfTotalNodes];
		
		this.lastTimeVisited = new double[noOfTotalNodes];
		this.forwardList = new LinkedList<Tuple<Message,Connection>>();

		for(int i = 0; i < noOfTotalNodes; i++) {
			for (int j = 0; j < noOfTotalNodes; j++) {                        
				intercontactInfo[i][j] = new InterContactInfo();
				routeInfo[i][j] = new RouteInfo();
				if (i == j) {
					routeInfo[i][j].delay = 0.0;
					routeInfo[i][j].variance = 0.0;					
				}
			}
		}
	}

	@Override
	public void changedConnection(Connection con) {
		if (con.isUp()) {                
			DTNHost otherHost = con.getOtherNode(getHost());

			lastTimeVisited[otherHost.getAddress()] = SimClock.getTime(); 

			/* Find pair-wise meeting contact with this host */
			for (int j = 0 ; j < noOfTotalNodes; j++) {     
				/* ij -> ik happens i: this node, j: neighbor, k: just visited */
				if (j == otherHost.getAddress()) continue; /* not the current node it meets */
				
				intercontactInfo[j][otherHost.getAddress()].addSample(lastTimeVisited[j], SimClock.getTime());
			}			
          
			/* for each destination, dest w, compute c(ij-->w) */
			for(int dest = 0; dest < noOfTotalNodes; dest++)				
                            routeInfo[otherHost.getAddress()][dest] = getBestRouteToDestination(otherHost, dest);

			// ACK messages
			MulticopyContactRouter otherRouter = (MulticopyContactRouter)otherHost.getRouter();

			for(Message m: otherRouter.ACKedMessages)
			{
				if (!this.ACKedMessages.contains(m) )
				{
					this.ACKedMessages.add(m);
				}                         

				if (this.getMessage(m.getId()) != null)
					this.deleteMessage(m.getId(), false);
			}
			
			// Compute number of copies to forward               
			List<Message> del = new ArrayList<Message>();

			for (Message m : getMessageCollection())
			{
				if (otherRouter.isDeliveredMessage(m.getId())) {
					del.add(m);
					continue;
				}
				
				// Check if the message is there or visited                    
				if (m.getHops().contains(otherHost) || otherRouter.getMessage(m.getId()) != null) continue; // the message already there

				ContactMessage cm = (ContactMessage)m;                    

				/*
                                double delay = routeInfo[otherHost.getAddress()][cm.getTo().getAddress()].delay;
                                double variance = routeInfo[otherHost.getAddress()][cm.getTo().getAddress()].variance;
                                double prob = assignFraction(delay, variance, cm.deadline - SimClock.getTime());        
                                
                                if (prob > cm.deliveryProb)
                                {
                                    ContactMessage cmnew = cm.replicate();
                                    cmnew.deliveryProb = prob;
                                    cm.deliveryProb = prob;                                    
                                    System.out.println(getHost() + " forwards " + cm + "to " +  otherHost + "old " + cm.deliveryProb + " new "+ prob);
                                    forwardList.add(new Tuple<Message, Connection>(cmnew, con));
                                }
                                */
                                
                                
				int copies = findNumberOfCopiesToForward(otherHost, cm);                        

				if (copies > 0 && copies < cm.copiesLeft)
				{
					log(getHost() + " transfering " + cm + " to " + otherHost + " with copies " + copies);
					ContactMessage copy = cm.replicate();
					copy.copiesLeft = copies;
                                        cm.copiesLeft = cm.copiesLeft - copies;
					forwardList.add(new Tuple<Message, Connection>(copy,con));
				}
                                 
			}
			
			for(Message m : del)
				deleteMessage(m.getId(), true);
			
		} else {
			for (Iterator<Tuple<Message, Connection>> iterator = forwardList.iterator(); iterator.hasNext();) {
				Tuple<Message, Connection> entry = iterator.next();
				if (entry.getValue().equals(con)) 
					iterator.remove();
			}
		}
	}

	void showContacts(Map<PeerEvent, PeerEventEntry> map, String name)
	{
		log(name);
		for (Map.Entry<PeerEvent, PeerEventEntry> e : map.entrySet())
			log(e.getKey() + " : " + e.getValue());
	}       

	/**
     * Schedules messages originated from this very node to some destination 
	 */

	@Override
	public void update() {            
		super.update();
		if (!canStartTransfer() || isTransferring()) {
			return; // nothing to transfer or is currently transferring 
		}
		// try messages that could be delivered to final recipient
		if (exchangeDeliverableMessages() != null) {
			return;
		}

		Collections.sort(forwardList, new TupleComparator());
		tryMessagesForConnected(forwardList);	// try to send messages		
	}

	@Override 
	public boolean createNewMessage(DTNHost to, String id, int size,
			int responseSize) {		

		makeRoomForNewMessage(size);		
		ContactMessage cm = new ContactMessage(getHost(), to, id, size, messageCopies);
		cm.setResponseSize(responseSize);
		cm.deadline = SimClock.getTime() + deadline;
		addToMessages(cm, true);

		return true;
	}
        
    @Override
        protected void transferAborted(Connection con)
        {
            ContactMessage m = (ContactMessage)con.getMessage();
            ContactMessage my_msg = (ContactMessage)this.getMessage(m.getId());
            if(my_msg == null) return;
            my_msg.copiesLeft += m.copiesLeft;
        }
        
	@Override
	protected void transferDone(Connection con) {            
		ContactMessage m = (ContactMessage)con.getMessage();            
		DTNHost other = con.getOtherNode(getHost());

		//		System.out.println("@@" + m.getId() + " " + m.getTo().getAddress() + " " + getHost().getAddress()  + "-" + con.getOtherNode(getHost()).getAddress() + " " + SimClock.getTime());
		log(getHost() + " transfer done " + m + " to " + other);

		ContactMessage my_msg = (ContactMessage)this.getMessage(m.getId());

		if (my_msg == null)
			return;

		for(Tuple<Message, Connection> forward: forwardList)
			if (forward.getKey().getId().equals(my_msg.getId()) && forward.getValue().equals(con)) {
				forwardList.remove(forward);
				break;
			}

		
		if (m.getTo().equals(other)) {
			deleteMessage(my_msg.getId(), false);
			ACKedMessages.add(m);
			return;
		}

		//my_msg.copiesLeft -= m.copiesLeft;                             
                
                if(my_msg.copiesLeft <=0)
                {
                    System.exit(1);
                }
                
		//if (my_msg.copiesLeft <= 0)
		//	deleteMessage(my_msg.getId(), true);
	}

	private RouteInfo getBestRouteToDestination(DTNHost neighbor, int destination)
	{
		if (neighbor.getAddress() == destination || getHost().getAddress() == destination)
                {
                    DTNHost dest = (neighbor.getAddress() == destination) ? neighbor : getHost();
                    List<DTNHost> hops = new ArrayList<DTNHost>();
                    hops.add(dest);
                    return new RouteInfo(0.0,0.0,(byte)0, hops);
                }
                
                MulticopyContactRouter otherRouter = (MulticopyContactRouter)neighbor.getRouter();

		int i = getHost().getAddress();

		double best_metric = -1.0;
		double best_delay = 0;
		double best_variance = 0;
                byte best_hopCount = 0;
                List<DTNHost> best_hops = new ArrayList<DTNHost>();
                        
		for(int k = 0; k < noOfTotalNodes; k++)
		{		
			if (k == i) continue;                               

			InterContactInfo neighbor_intercontact = otherRouter.intercontactInfo[i][k];
			
			if (!neighbor_intercontact.isValid(SimClock.getTime()))
				continue;
		
			RouteInfo neighbor_route = otherRouter.routeInfo[k][destination];		
			
			if (neighbor_route.delay < 0 || neighbor_route.hops.contains(getHost()))
				continue;
		
			double delay = neighbor_intercontact.delay + neighbor_route.delay;
			double variance = neighbor_intercontact.variance + neighbor_route.variance;
                        byte hopCount = (byte) (1 + neighbor_route.hopCount);
                        List<DTNHost> hops = neighbor_route.hops;
                        
			double metric = delay + 3.0 * Math.sqrt(variance);

			//if (best_metric < 0 || hopCount < best_hopCount || (hopCount == best_hopCount && metric < best_metric))
                        if (best_metric < 0 || metric < best_metric)
			{
				best_metric = metric;                                
				best_delay = delay;
				best_variance = variance;
                                best_hopCount = hopCount;
                                best_hops.clear();
                                best_hops.add(getHost());
                                for(DTNHost d : hops)
                                    best_hops.add(d);
			}
		}

		if (best_metric > 0)
			return new RouteInfo(best_delay, best_variance, best_hopCount, best_hops);
		else
			return new RouteInfo();
	}
	
	private double assignFraction(double delay, double variance, double deadline) {
		//log("Delay : " + delay + " variance " + Math.sqrt(variance));
		/*            if (Math.sqrt(variance) >= delay) 
                return 0.0;
            return Math.exp(-1.0 * variance/(delay * delay));*/

		double prob = Phi((deadline - delay)/Math.sqrt(variance)) - Phi(-1.0 * delay / Math.sqrt(variance)) ; /// (1 - Phi(-1.0 * delay / Math.sqrt(variance))) ;
                //System.out.println("===========Prob " + prob);
                return prob;
                //return (prob > 0.8)? 1: prob;
                /*
                if (deadline > delay)
			return Phi((deadline - delay) / Math.sqrt(variance));
		else
			return 1.0 - Phi((delay - deadline) / Math.sqrt(variance));
                */ 
	}       

	private int findNumberOfCopiesToForward(DTNHost other, ContactMessage m)
	{         
		ArrayList<Tuple<Integer, Tuple<Integer, Double>>> forward_list = new ArrayList<Tuple<Integer, Tuple<Integer, Double>>>();
		
		if (m.getHops().contains(other.getAddress())) return 0;
		
                int copies = 0;
                int totalCopies = 0;
                
		RouteInfo info = null;
                
                info = getBestRouteToDestination(other, m.getTo().getAddress());
                if (info.delay > 0)
                {                    
                    double metric = info.delay + 3.0 * Math.sqrt(info.variance);                    
                    copies = (int)Math.floor(m.copiesLeft * assignFraction(info.delay, info.variance, m.deadline - SimClock.getTime()));                    
                    totalCopies += copies;
                   // System.out.println("Decided: " + copies);
                    forward_list.add(new Tuple<Integer, Tuple<Integer, Double>>(other.getAddress(), new Tuple<Integer, Double>(copies, metric)));                    
                }
		
                /*
                MulticopyContactRouter otherRouter = (MulticopyContactRouter)other.getRouter();                
                
                double best_metric = -1.0;
                
                for(int i = 0 ; i < noOfTotalNodes; i++)
                {
                    if (i == getHost().getAddress()) continue;
                    
                    double delay = otherRouter.routeInfo[i][m.getTo().getAddress()].delay;
                    double variance = otherRouter.routeInfo[i][m.getTo().getAddress()].variance;
                    double metric = delay + 3.0 * Math.sqrt(variance);
                    
                    if (delay > 0 && (best_metric < 0 || best_metric > metric))
                    {
                        info = otherRouter.routeInfo[i][m.getTo().getAddress()];
                    }                        
                }
                
                if (info != null && info.delay > 0)
                {
                    copies = (int)Math.ceil(m.copiesLeft * assignFraction(info.delay, info.variance, m.deadline - SimClock.getTime()));
                    //copies = Math.max(copies, info.hopCount);                                    
                }
                */

		for (int i = 0; i < noOfTotalNodes; i++) {
			if (m.getHops().contains(i) || i == other.getAddress()) 
				continue;
			
			InterContactInfo intercontact = intercontactInfo[other.getAddress()][i]; 
			
			if (!intercontact.isValid(SimClock.getTime()))
				continue;
			
			RouteInfo route = routeInfo[other.getAddress()][m.getTo().getAddress()];

			if (route.delay < 0)
				continue;
				
			double delay = intercontact.delay + route.delay;
			double variance = intercontact.variance + route.variance;

                        //if (route.hopCount <= otherHopCount)
                        copies = (int)Math.floor(m.copiesLeft * assignFraction(delay, variance, m.deadline - SimClock.getTime()));
                        totalCopies += copies;    
			//copies = Math.max(copies, route.hopCount);                        
                        
                        double metric = delay + 3.0 * Math.sqrt(variance);
			forward_list.add(new Tuple<Integer, Tuple<Integer, Double>>(i, new Tuple<Integer, Double>(copies, metric)));
		} 

		Collections.sort(forward_list, new ForwardListTupleComparator());     
		
		int remaining_copies = m.copiesLeft - 1;
		int assigned_copies = 0;
                double factor = 1.0;
                
                //if (totalCopies > 0 && totalCopies < remaining_copies)
                //    factor = remaining_copies / totalCopies;
                                  

		for (Tuple<Integer, Tuple<Integer, Double>> t: forward_list) {
			if (remaining_copies <= 0)
				break;

                        int tmpCopies = (int)(factor * t.getValue().getKey());        
			if (t.getKey() == other.getAddress()) 
				assigned_copies = Math.min(remaining_copies, tmpCopies);    

			remaining_copies -= tmpCopies;                
		}           

		if (remaining_copies > 0)
			assigned_copies += remaining_copies / 2;

                //System.out.println("But: " + assigned_copies);
                
		//if (assigned_copies > 0)
		//	System.out.println("Computing " + getHost().getAddress() + " -> " + other.getAddress() + " forward list : " + forward_list);
                               
		return assigned_copies;
	}

	private class ForwardListTupleComparator implements Comparator                 
	<Tuple<Integer, Tuple<Integer, Double>>> {

		public int compare(Tuple<Integer, Tuple<Integer, Double>> tuple1,
				Tuple<Integer, Tuple<Integer, Double>> tuple2) {  

			return (int)(tuple1.getValue().getValue() - tuple2.getValue().getValue());
		}            
	}

	/**
	 * Comparator for Message-Connection-Tuples that orders the tuples by
	 * their delivery probability by the host on the other side of the 
	 * connection (GRTRMax)
	 */
	private class TupleComparator implements Comparator                 
	<Tuple<Message, Connection>> {

		public int compare(Tuple<Message, Connection> tuple1,
				Tuple<Message, Connection> tuple2) {                

			ContactMessage m1 = (ContactMessage)tuple1.getKey();
			ContactMessage m2 = (ContactMessage)tuple2.getKey();

			return (int)(m1.deadline - m2.deadline);
		}            
	}

	@Override
	public RoutingInfo getRoutingInfo() {        
		DTNHost hi = getHost();
		RoutingInfo top = super.getRoutingInfo();
		RoutingInfo ri1 = new RoutingInfo("Contacts");

		for (int j = 0; j < noOfTotalNodes; j++)
			for (int k = 0; k < noOfTotalNodes; k++) {

				InterContactInfo intercontact = intercontactInfo[j][k];
				if (j == k || !intercontact.isValid(SimClock.getTime())) continue;

				ri1.addMoreInfo(new RoutingInfo(String.format("(%s,%s) -> (%s,%s) = %.3f, %.3f", hi, j, hi, k, intercontact.delay, intercontact.variance)));     
			}
		//top.addMoreInfo(ri1);

		RoutingInfo ri2 = new RoutingInfo("Routes");

		for (int j = 0; j < noOfTotalNodes; j++)
			for(int w = 0; w < noOfTotalNodes; w++) {
				
				RouteInfo route = routeInfo[j][w];

				if (route.delay <= 0) continue;
                        
				ri2.addMoreInfo(new RoutingInfo(String.format("(%d,%d) -> %s: %.3f , %.3f: %s", hi.getAddress(), j, w, route.delay, route.variance, route.hops)));      
			}   

		top.addMoreInfo(ri2);
		return top;            
	}

	@Override
	public MessageRouter replicate() {
		MulticopyContactRouter r = new MulticopyContactRouter(this);
		return r;
	}

	private double Phi(double x)
	{
		final double b1 =  0.319381530;
		final double b2 = -0.356563782;
		final double b3 =  1.781477937;
		final double b4 = -1.821255978;
		final double b5 =  1.330274429;
		final double p  =  0.2316419;
		final double c  =  0.39894228;

		if(x >= 0.0) {
			double t = 1.0 / ( 1.0 + p * x );
			return (1.0 - c * Math.exp( -x * x / 2.0 ) * t *
					( t *( t * ( t * ( t * b5 + b4 ) + b3 ) + b2 ) + b1 ));
		}
		else {
			double t = 1.0 / ( 1.0 - p * x );
			return ( c * Math.exp( -x * x / 2.0 ) * t *
					( t *( t * ( t * ( t * b5 + b4 ) + b3 ) + b2 ) + b1 ));
		}
	}  

	private static void log(String msg)
	{
		if (debug)
			System.out.println(SimClock.getTime() + " " + msg);
	}

	private class ContactMessage extends Message{
		public int copiesLeft;
		public double deadline;
                    
                public double deliveryProb;
                
		public ContactMessage(DTNHost from, DTNHost to, String id, int size, int copies) {
			super(from, to, id, size, msgTtl);			
			copiesLeft = copies;		
                        deliveryProb = 0.0;
		}

		@Override
		public ContactMessage replicate() {
			ContactMessage m = new ContactMessage(getFrom(),getTo(),getId(),getSize(), copiesLeft);
			m.deadline = deadline;
                        m.deliveryProb = deliveryProb;

			m.copyFrom(this);                
			return m;
		}

		@Override
		public String toString() {			
			return getId()+ "(" + getFrom() + "=>" + getTo() + ")[" + copiesLeft + "] " + getHops();
		}
	}        
}
