/* 
 * Copyright 2007 TKK/Netlab
 * Released under GPLv3. See LICENSE.txt for details. 
 *
 * @author Mehedi Bakht
 */

package routing;

import java.util.ArrayList;
import java.util.List;

import core.Connection;
import core.DTNHost;
import core.Message;
import core.Settings;
import core.SimClock;
/**
 * Implementation of Spray and Foucs router as depicted in 
 * <I>Spray and Wait: An Efficient Routing Scheme for Intermittently
 * Connected Mobile Networks</I> by Thrasyvoulos Spyropoulus et al.
 *
 */
public class SprayAndFocusRouterMehedi extends ActiveRouter {
	/** identifier for the intial number of copies setting ({@value})*/ 
	public static final String NROF_COPIES = "nrofCopies";
	/** identifier for the binary-mode setting ({@value})*/ 
	public static final String BINARY_MODE = "binaryMode";
	/** SprayAndWait router's settins namespace ({@value})*/ 
	public static final String SPRAYANDFOCUS_NS = "SprayAndFocusRouter";
	
        public static final String ALPHA = "alpha";
        public static final String UPDATE_POPINTERVAL = "updatePOPInterval";
	
        private int initialNrofCopies;
	private boolean isBinary;
        /** Current window counter */
        private double windowCounter = 0;
        /** Alpha */
        private double alpha;
        
        /** Future time to update popCounter */
        private double updateInterval;
        private double timeToUpdate = 0;
        
        /** Popularity counter */
        private double popCounter = 0;

	public SprayAndFocusRouterMehedi(Settings s) {
		super(s);
		Settings snfSettings = new Settings(SPRAYANDFOCUS_NS);
		
		initialNrofCopies = snfSettings.getInt(NROF_COPIES);
		isBinary = snfSettings.getBoolean( BINARY_MODE);
                alpha = snfSettings.getDouble(ALPHA);
                updateInterval = snfSettings.getInt(UPDATE_POPINTERVAL);                
	}
	
	/**
	 * Copy constructor.
	 * @param r The router prototype where setting values are copied from
	 */
	protected SprayAndFocusRouterMehedi(SprayAndFocusRouterMehedi r) {
		super(r);
		this.initialNrofCopies = r.initialNrofCopies;
		this.isBinary = r.isBinary;
                this.alpha = r.alpha;
                this.updateInterval = r.updateInterval;
	}
	
	@Override
	public int receiveMessage(Message m, DTNHost from) {
		assert m instanceof SAFMessage : this.getClass().getSimpleName() + 
			" accepts only SAFMessages";
		return super.receiveMessage(m, from);
	}
	
	@Override
	public Message messageTransferred(String id, DTNHost from) {
		SAFMessage msg = (SAFMessage)super.messageTransferred(id, from);
		
		if (isBinary) {
			// in binary S'n'W the receiving node gets ceil(n/2) copies
			msg.nrofCopiesLeft = (int)Math.ceil(msg.nrofCopiesLeft/2.0);
		}
		else {
			// in standard S'n'W the receiving node gets only single copy
			msg.nrofCopiesLeft = 1;
		}
		
		return msg;
	}
	
	@Override 
	public boolean createNewMessage(DTNHost to, String id, int size,
			int responseSize) {
		makeRoomForNewMessage(size);
		
		SAFMessage m = new SAFMessage(getHost(), to, id, size);
		m.setResponseSize(responseSize);
		addToMessages(m, true);
				
		return true;
	}
	
	@Override
	public void update() {
		super.update();
              
                // Update popularity counter if need be
                if (SimClock.getTime() >= timeToUpdate) {
                    popCounter = alpha * windowCounter + (1-alpha) * popCounter;
                    windowCounter = 0;
                    timeToUpdate = SimClock.getTime() + updateInterval;
                }
                
                // Check for 1-hop interference
                // if (oneHopInterference()) {
                //    return;
                //}
                
		if (!canStartTransfer() || isTransferring()) {
			return; // nothing to transfer or is currently transferring 
		}

		// try messages that could be delivered to final recipient
		if (exchangeDeliverableMessages() != null) {
			return;
		}
		
		// create a list of SAFMessages that have copies left to distribute
		List<Message> copiesLeft = getMessagesWithCopiesLeft();
                
               	// try to send those messages
		this.tryAllMessagesToAllConnections(copiesLeft, getConnections());		
                
                trySingleCopyMessages();
	}
        
        private void trySingleCopyMessages() {
            // create a list of SAFMessages that have just a single copy left
            List<Message> singleCopyMsgs = getMessagesWithSingleCopy();
            double max = 0;
            Connection mostPopularConnection = null;
            // Collect popularity information of all connected hosts
            for (Connection con : getConnections()) {
			DTNHost other = con.getOtherNode(getHost());
			SprayAndFocusRouterMehedi othRouter = (SprayAndFocusRouterMehedi)other.getRouter();
                        
                        double popularity = othRouter.getPopCounter();
                        if(popularity > max) {
                                max = popularity;
                                mostPopularConnection = con;
                        }
            }
            if(mostPopularConnection == null || max <= getPopCounter()) {
                return;
            }
            
            this.tryAllMessages(mostPopularConnection, singleCopyMsgs);
        }

        
        public void changedConnection(Connection con) {
		if (con.isUp()) {
                       
                   /** Update current window counter */
                    windowCounter++;  
                }
	}
	/**
	 * Creates and returns a list of messages this router is currently
	 * carrying and still has copies left to distribute (nrof copies > 1).
	 * @return A list of messages that have copies left
	 */
	private List<Message> getMessagesWithCopiesLeft() {
		List<Message> list = new ArrayList<Message>();

		for (Message m : getMessageCollection()) {
			if (((SAFMessage) m).nrofCopiesLeft > 1) {
				list.add(m);
			}
		}
		
		return list;
	}
        
        
	/**
	 * Creates and returns a list of messages this router is currently
	 * carrying and that has just a single copy (nrof copies = 1).
	 * @return A list of messages that have copies left
	 */
	private List<Message> getMessagesWithSingleCopy() {
		List<Message> list = new ArrayList<Message>();

		for (Message m : getMessageCollection()) {
			if (((SAFMessage) m).nrofCopiesLeft == 1) {
				list.add(m);
			}
		}
		
		return list;
	}
	
	/**
	 * Called just before a transfer is finalized (by 
	 * {@link ActiveRouter#update()}).
	 * Reduces the number of copies we have left for a message. 
	 * In binary Spray and Wait, sending host is left with floor(n/2) copies,
	 * but in standard mode, nrof copies left is reduced by one. 
	 */
	@Override
	protected void transferDone(Connection con) {
		String msgId = con.getMessage().getId();
		// get this router's copy of the message
		SAFMessage msg = (SAFMessage)getMessage(msgId);

		if (msg == null) { // message has been dropped from the buffer after..
			return; // ..start of transfer -> no need to reduce amount of copies
		}
		
		// reduce the amount of copies left
		if (isBinary) { 
			msg.nrofCopiesLeft /= 2;
		}
		else {
			msg.nrofCopiesLeft--;
		}
	}
	
	@Override
	public SprayAndFocusRouterMehedi replicate() {
		return new SprayAndFocusRouterMehedi(this);
	}

       
        public double getPopCounter() {
            return this.popCounter;
        }
        
	/**
	 * Spray and Focus message
	 */
	private class SAFMessage extends Message{
		private int nrofCopiesLeft;
		
		public SAFMessage(DTNHost from, DTNHost to, String id, int size) {
			super(from, to, id, size, msgTtl);
			this.nrofCopiesLeft = initialNrofCopies;
		}

		@Override
		public SAFMessage replicate() {
			SAFMessage m = new SAFMessage(getFrom(),getTo(),getId(),getSize());
			m.copyFrom(this);
			m.nrofCopiesLeft = nrofCopiesLeft;
			return m;
		}
		
		public String toString() {
			return super.toString() + " (" + this.nrofCopiesLeft + 
			" copies)";
		}
	}
}



