/* 
 * Copyright 2007 TKK/Netlab
 * Released under GPLv3. See LICENSE.txt for details. 
 */
package routing;

import java.util.ArrayList;
import java.util.List;

import core.Connection;
import core.DTNHost;
import core.Message;
import core.Settings;

/**
 * Implementation of Spray and wait router as depicted in 
 * <I>Spray and Wait: An Efficient Routing Scheme for Intermittently
 * Connected Mobile Networks</I> by Thrasyvoulos Spyropoulus et al.
 *
 */
public class SprayAndWaitRouter extends ActiveRouter {
	/** identifier for the intial number of copies setting ({@value})*/ 
	public static final String NROF_COPIES = "nrofCopies";
	/** identifier for the binary-mode setting ({@value})*/ 
	public static final String BINARY_MODE = "binaryMode";
	/** SprayAndWait router's settins namespace ({@value})*/ 
	public static final String SPRAYANDWAIT_NS = "SprayAndWaitRouter";
	
	private int initialNrofCopies;
	private boolean isBinary;

	public SprayAndWaitRouter(Settings s) {
		super(s);
		Settings snwSettings = new Settings(SPRAYANDWAIT_NS);
		
		initialNrofCopies = snwSettings.getInt(NROF_COPIES);
		isBinary = snwSettings.getBoolean( BINARY_MODE);
	}
	
	/**
	 * Copyconstructor.
	 * @param r The router prototype where setting values are copied from
	 */
	protected SprayAndWaitRouter(SprayAndWaitRouter r) {
		super(r);
		this.initialNrofCopies = r.initialNrofCopies;
		this.isBinary = r.isBinary;
	}
	
	@Override
	public int receiveMessage(Message m, DTNHost from) {
		assert m instanceof SAWMessage : this.getClass().getSimpleName() + 
			" accepts only SAWMessages";
		return super.receiveMessage(m, from);
	}
	
	@Override
	public Message messageTransferred(String id, DTNHost from) {
		SAWMessage msg = (SAWMessage)super.messageTransferred(id, from);
		
		if (isBinary) {
			// in binary S'n'W the receiving node gets ceil(n/2) copies
			msg.nrofCopiesLeft = (int)Math.ceil(msg.nrofCopiesLeft/2.0);
		}
		else {
			// in standard S'n'W the receiving node gets only single copy
			msg.nrofCopiesLeft = 1;
		}
		
		return msg;
	}
	
	@Override 
	public boolean createNewMessage(DTNHost to, String id, int size,
			int responseSize) {
		makeRoomForNewMessage(size);
		
		SAWMessage m = new SAWMessage(getHost(), to, id, size);
		m.setResponseSize(responseSize);
		addToMessages(m, true);
				
		return true;
	}
	
	@Override
	public void update() {
		super.update();
		if (!canStartTransfer() || isTransferring()) {
			return; // nothing to transfer or is currently transferring 
		}

		// try messages that could be delivered to final recipient
		if (exchangeDeliverableMessages() != null) {
			return;
		}
		
		// create a list of SAWMessages that have copies left to distribute
		List<Message> copiesLeft = getMessagesWithCopiesLeft();
		
		// try to send those messages
		this.tryAllMessagesToAllConnections(copiesLeft, getConnections());		
	}
	
	/**
	 * Creates and returns a list of messages this router is currently
	 * carrying and still has copies left to distribute (nrof copies > 1).
	 * @return A list of messages that have copies left
	 */
	private List<Message> getMessagesWithCopiesLeft() {
		List<Message> list = new ArrayList<Message>();

		for (Message m : getMessageCollection()) {
			if (((SAWMessage)m).nrofCopiesLeft > 1) {
				list.add(m);
			}
		}
		
		return list;
	}
	
	/**
	 * Called just before a transfer is finalized (by 
	 * {@link ActiveRouter#update()}).
	 * Reduces the number of copies we have left for a message. 
	 * In binary Spray and Wait, sending host is left with floor(n/2) copies,
	 * but in standard mode, nrof copies left is reduced by one. 
	 */
	@Override
	protected void transferDone(Connection con) {
		String msgId = con.getMessage().getId();
		// get this router's copy of the message
		SAWMessage msg = (SAWMessage)getMessage(msgId);

		if (msg == null) { // message has been dropped from the buffer after..
			return; // ..start of transfer -> no need to reduce amount of copies
		}
		
		// reduce the amount of copies left
		if (isBinary) { 
			msg.nrofCopiesLeft /= 2;
		}
		else {
			msg.nrofCopiesLeft--;
		}
	}
	
	@Override
	public SprayAndWaitRouter replicate() {
		return new SprayAndWaitRouter(this);
	}

	/**
	 * Spray and Wait message
	 */
	private class SAWMessage extends Message{
		private int nrofCopiesLeft;
		
		public SAWMessage(DTNHost from, DTNHost to, String id, int size) {
			super(from, to, id, size, msgTtl);
			this.nrofCopiesLeft = initialNrofCopies;
		}

		@Override
		public SAWMessage replicate() {
			SAWMessage m = new SAWMessage(getFrom(),getTo(),getId(),getSize());
			m.copyFrom(this);
			m.nrofCopiesLeft = nrofCopiesLeft;
			return m;
		}
		
		public String toString() {
			return super.toString() + " (" + this.nrofCopiesLeft + 
			" copies)";
		}
	}
}



