/* 
 * Copyright 2007 TKK/Netlab
 * Released under GPLv3. See LICENSE.txt for details. 
 */
package routing.maxprop;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import core.SimClock;

/**
 * Class for storing and manipulating the meeting probabilities for the MaxProp
 * router module.
 */
public class MeetingProbabilitySetMOD {	
        
        /** meeting probabilities (probability that the next node one meets is X) */
	protected Map<Integer, Double> probs;
	/** the time when this MPS was last updated */
	protected double lastUpdateTime;
        
        private Map<Integer, Double> meets;	
	/**
	 * Constructor. Creates a probability set with empty node-probability
	 * mapping.
	 */
	public MeetingProbabilitySetMOD() {
		this.probs = new HashMap<Integer, Double>();
                this.meets = new HashMap<Integer, Double>();
		this.lastUpdateTime = 0;
	}
	
	/**
	 * Constructor. Creates a probability set with equal probability for
	 * all the given node indexes.
	 */
	public MeetingProbabilitySetMOD(List<Integer> initiallyKnownNodes) {
		this();
		double prob = 1.0/initiallyKnownNodes.size();
		for (Integer i : initiallyKnownNodes) {
			this.meets.put(i, 1.0);
                        this.probs.put(i, Math.pow(Math.E, (10.0)));
		}
	}
	
	/**
	 * Updates meeting probability for the given node index.
	 * <PRE> P(b) = P(b)_old + 1
	 * Normalize{P}</PRE> 
	 * I.e., The probability of the given node index is increased by one and
	 * then all the probabilities are normalized so that their sum equals to 1.
	 * @param index The node index to update the probability for
	 */
	public void updateMeetingProbFor(Integer index) {
		this.lastUpdateTime = SimClock.getTime();
		
		if (probs.size() == 0) { // first entry
			meets.put(index, 1.0);
                        probs.put(index, Math.pow(Math.E, 10.0));
			return;
		}
		
		double newValue = getMeetFor(index) + 1;
		meets.put(index, newValue);
		probs.put(index, Math.pow(Math.E, Math.pow(10.0/newValue, 2)));
		
                //System.out.println("Setting prob for " + index);
                /* now the sum of all entries is 2;
		 * normalize to one by dividing all the entries by 2 */ 
		//for (Map.Entry<Integer, Double> entry : probs.entrySet()) {
		//	entry.setValue(entry.getValue() / 2.0);
	        //}
	}
	
        public double getMeetFor(Integer index) {
		if (meets.containsKey(index)) {
			return meets.get(index);
		}
		else {
			/* the node with the given index has not been met */
			return 1.0;
		}
	}
	/**
	 * Returns the current delivery probability value for the given node index 
	 * @param index The index of the node to look the P for
	 * @return the current delivery probability value
	 */
	public double getProbFor(Integer index) {
		if (probs.containsKey(index)) {
			return probs.get(index);
		}
		else {
			/* the node with the given index has not been met */
			return 0.0;
		}
	}
	
	/**
	 * Returns a reference to the probability map of this probability set
	 * @return a reference to the probability map of this probability set
	 */
	public Map<Integer, Double> getAllProbs() {
		return this.probs;
	}
	
	/**
	 * Returns the time when this probability set was last updated
	 * @return the time when this probability set was last updated
	 */
	public double getLastUpdateTime() {
		return this.lastUpdateTime;
	}
	
	/**
	 * Returns a deep copy of the probability set
	 * @return a deep copy of the probability set
	 */
	public MeetingProbabilitySetMOD replicate() {
		MeetingProbabilitySetMOD replicate = new MeetingProbabilitySetMOD();
		// do a deep copy
		for (Map.Entry<Integer, Double> e : probs.entrySet()) {
			replicate.probs.put(e.getKey(), e.getValue().doubleValue());
		}
		
                for (Map.Entry<Integer, Double> e : meets.entrySet()) {
			replicate.meets.put(e.getKey(), e.getValue().doubleValue());
		}
		replicate.lastUpdateTime = this.lastUpdateTime;
		return replicate;
	}
	
	/**
	 * Returns a String presentation of the probabilities
	 * @return a String presentation of the probabilities
	 */
	public String toString() {
		return "probs: " +	this.probs.toString();
	}
}